import time
import weakref

import numpy as np
import wgpu.backends.rs

from .. import Renderer, RenderFunctionRegistry
from ...linalg import Matrix4, Vector3
from ...objects import WorldObject
from ...cameras import Camera
from ...resources import Buffer, TextureView
from ...utils import array_from_shadertype

from .postprocessing import RenderTexture, SSAAPostProcessingStep


# Definition uniform struct with standard info related to transforms,
# provided to each shader as uniform at slot 0.
# todo: a combined transform would be nice too, for performance
stdinfo_uniform_type = dict(
    cam_transform=("float32", (4, 4)),
    cam_transform_inv=("float32", (4, 4)),
    projection_transform=("float32", (4, 4)),
    projection_transform_inv=("float32", (4, 4)),
    physical_size=("float32", 2),
    logical_size=("float32", 2),
)


registry = RenderFunctionRegistry()

visibility_all = (
    wgpu.ShaderStage.VERTEX | wgpu.ShaderStage.FRAGMENT | wgpu.ShaderStage.COMPUTE
)

# Alternative texture formats that we support by padding channels as needed.
# Maps virtual_format -> (wgpu_format, pad_value, nbytes)
ALTTEXFORMAT = {
    "rgb8snorm": ("rgba8snorm", 127, 1),
    "rgb8unorm": ("rgba8unorm", 255, 1),
    "rgb8sint": ("rgba8sint", 127, 1),
    "rgb8uint": ("rgba8uint", 255, 1),
    "rgb16sint": ("rgba16sint", 32767, 2),
    "rgb16uint": ("rgba16uint", 65535, 2),
    "rgb32sint": ("rgba32sint", 2147483647, 4),
    "rgb32uint": ("rgba32uint", 4294967295, 4),
    "rgb16float": ("rgba16float", 1, 2),
    "rgb32float": ("rgba32float", 1, 4),
}


def register_wgpu_render_function(wobject_cls, material_cls):
    """Decorator to register a WGPU render function."""

    def _register_wgpu_renderer(f):
        registry.register(wobject_cls, material_cls, f)
        return f

    return _register_wgpu_renderer


class RenderInfo:
    """The type of object passed to each wgpu render function together
    with the world object. Contains stdinfo buffer for now. In time
    will probably also include lights etc.
    """

    def __init__(self, *, stdinfo_uniform):
        self.stdinfo_uniform = stdinfo_uniform


class WgpuRenderer(Renderer):
    """A renderer that renders to a surface."""

    def __init__(self, canvas=None, size=None, pixel_ratio=None, show_fps=False):

        # Check and normalize inputs
        if not (canvas is None or isinstance(canvas, wgpu.gui.WgpuCanvasBase)):
            raise TypeError("The given canvas must be None or a wgpu Canvas.")
        self._canvas = canvas
        if not (size is None or (isinstance(size, (tuple, list)) and len(size) == 2)):
            raise TypeError("The given size must be None or a 2-tuple.")
        self._logical_size = (float(size[0]), float(size[1])) if size else None
        self.pixel_ratio = pixel_ratio
        self._show_fps = bool(show_fps)

        # Do we have enough info?
        if canvas is None and size is None:
            raise ValueError("WgpuRenderer needs either a canvas or a size.")

        # Create adapter and device objects
        self._adapter = wgpu.request_adapter(
            canvas=canvas, power_preference="high-performance"
        )
        self._device = self._adapter.request_device(
            non_guaranteed_features=[], non_guaranteed_limits={}
        )

        self._shader_cache = {}

        # Prepare render targets. These are placeholders to set during
        # each renderpass, intended to keep together the texture-view
        # object, its size, and its format. The final (canvas) texture
        # has bgra format, because that's what hardware seems to like,
        # and srgb for perseptive color mapping.
        self._canvas_texture = RenderTexture(wgpu.TextureFormat.bgra8unorm_srgb)
        self._depth_texture = RenderTexture(wgpu.TextureFormat.depth32float)
        # The pick texture has 4 channels, object id, and then 3 more, e.g.
        # the instance nr, vertex nr and weights.
        self._pick_texture = RenderTexture(wgpu.TextureFormat.rgba32sint)
        # We use one or two render textures (for 2+ steps, cycle between the 2)
        self._render_textures = []

        # Prepare post-processing steps. Users can append/insert more
        self._ssa_post_processing_step = SSAAPostProcessingStep(self._device)
        self._postfx = []

        # Prepare other properties
        self._msaa = 1  # todo: cannot set sample_count of render_pass yet

        # Initialize swapchain
        self._swap_chain = None
        if canvas is not None:
            self._swap_chain = canvas.configure_swap_chain(
                device=self._device,
                format=self._canvas_texture.format,
                usage=wgpu.TextureUsage.RENDER_ATTACHMENT,
            )

        # Initialize a small buffer to read pixel info into
        # Make it 256 bytes just in case (for bytes_per_row)
        self._pixel_info_buffer = self._device.create_buffer(
            size=256,
            usage=wgpu.BufferUsage.COPY_DST | wgpu.BufferUsage.MAP_READ,
        )

        # Keep track of object ids
        self._pick_map = weakref.WeakValueDictionary()

    @property
    def device(self):
        """A reference to the used wgpu device."""
        return self._device

    @property
    def postfx(self):
        """A list of post processing steps. Users can append
        PostProcessingStep objects here to do post-processing of the
        rendered image.

        Warning, this API is provisional.
        """
        # todo: is this the way to go?
        return self._postfx

    @property
    def pixel_ratio(self):
        """Configure the size of the render texture relative to the
        logical size. By default (value is None) the used pixel ratio
        follows the screens pixel ratio on high-res displays, and is 2
        otherwise.

        If the used pixel ratio causes the render texture to be larger
        than the physical size of the canvas, SSAA is applied, resulting
        in a smoother final image with less jagged edges. Alternatively,
        this value can be set to e.g. 0.5 to lower* the resolution (e.g.
        for performance during interaction).
        """
        return self._pixel_ratio

    @pixel_ratio.setter
    def pixel_ratio(self, value):
        if value is None:
            self._pixel_ratio = None
        elif isinstance(value, (int, float)):
            self._pixel_ratio = None if value <= 0 else float(value)
        else:
            raise TypeError(
                f"Rendered.pixel_ratio expected None or number, not {value}"
            )

    def render(self, scene: WorldObject, camera: Camera):
        """Main render method, called from the canvas."""

        device = self._device

        now = time.perf_counter()  # noqa
        if self._show_fps:
            if not hasattr(self, "_fps"):
                self._fps = now, now, 1
            elif now > self._fps[0] + 1:
                print(f"FPS: {self._fps[2]/(now - self._fps[0]):0.1f}")
                self._fps = now, now, 1
            else:
                self._fps = self._fps[0], now, self._fps[2] + 1

        # todo: support for alt render pipelines (object that renders to texture then renders that)
        # todo: also note that the fragment shader is (should be) optional
        #      (e.g. depth only passes like shadow mapping or z prepass)

        # Get logical size (as two floats). This size is constant throughout
        # all post-processing render passes.
        if self._logical_size:
            logical_size = self._logical_size
        else:
            logical_size = self._canvas.get_logical_size()

        # Determine the physical size of the render texture
        if self._canvas:
            canvas_size = self._canvas.get_physical_size()
            canvas_ratio = canvas_size[0] / logical_size[0]
            if self._pixel_ratio:
                pixel_ratio = self._pixel_ratio
            else:
                pixel_ratio = canvas_ratio
                if pixel_ratio <= 1:
                    pixel_ratio = 2.0  # use 2 on non-hidpi displays
        else:
            pixel_ratio = self._pixel_ratio if self._pixel_ratio else 2.0

        # Determine the physical size of the first and last render pass
        scene_size = tuple(max(1, int(pixel_ratio * x)) for x in logical_size)
        final_size = canvas_size if self._canvas else scene_size  # noqa

        # Create more render textures if needed, drop some if we can
        pp_steps = self._postfx
        render_textures = self._render_textures
        render_texture_count = min(2, len(pp_steps) + 1)
        while len(render_textures) < render_texture_count:
            render_textures.append(RenderTexture(wgpu.TextureFormat.rgba8unorm))
        while len(render_textures) > render_texture_count:
            render_textures.pop(-1)

        # Set the size of the textures (is a no-op if the size does not change)
        for t in render_textures:
            t.ensure_size(device, scene_size + (1,))
        self._depth_texture.ensure_size(device, scene_size + (1,))
        self._pick_texture.ensure_size(device, scene_size + (1,))

        # Ensure that matrices are up-to-date
        scene.update_matrix_world()
        camera.set_viewport_size(*logical_size)
        camera.update_matrix_world()  # camera may not be a member of the scene
        camera.update_projection_matrix()

        # Get the list of objects to render (visible and having a material)
        q = self.get_render_list(scene, camera)
        for wobject in q:
            self._pick_map[wobject.id] = wobject

        # Update stdinfo uniform buffer object that we'll use during this render call
        self._update_stdinfo_buffer(camera, scene_size, logical_size)

        # Ensure each wobject has pipeline info
        for wobject in q:
            self._ensure_up_to_date(wobject)

        # Filter out objects that we cannot render
        q = [wobject for wobject in q if wobject._wgpu_pipeline_objects is not None]

        # Render the scene graph (to the first texture)
        command_encoder = device.create_command_encoder()
        self._render_recording(command_encoder, q)
        command_buffers = [command_encoder.finish()]
        device.queue.submit(command_buffers)

        # Render each texture into the next
        for i in range(len(pp_steps)):
            step = pp_steps[i]
            step.render(
                render_textures[0],
                render_textures[1],
            )
            render_textures.insert(0, render_textures.pop(1))  # cycle

        # If we have a canvas, we render into it, applying SSAA if possible
        if self._canvas:
            with self._swap_chain as texture_view_target:
                self._canvas_texture.set_texture_view(texture_view_target)
                self._ssa_post_processing_step.render(
                    render_textures[0], self._canvas_texture
                )

    def _render_recording(self, command_encoder, q):

        # You might think that this is slow for large number of world
        # object. But it is actually pretty good. It does iterate over
        # all world objects, and over stuff in each object. But that's
        # it, really.
        # todo: we may be able to speed this up with render bundles though

        # ----- compute pipelines

        compute_pass = command_encoder.begin_compute_pass()

        for wobject in q:
            wgpu_data = wobject._wgpu_pipeline_objects
            for pinfo in wgpu_data["compute_pipelines"]:
                compute_pass.set_pipeline(pinfo["pipeline"])
                for bind_group_id, bind_group in enumerate(pinfo["bind_groups"]):
                    compute_pass.set_bind_group(
                        bind_group_id, bind_group, [], 0, 999999
                    )
                compute_pass.dispatch(*pinfo["index_args"])

        compute_pass.end_pass()

        # ----- render pipelines rendering to the default target

        assert self._render_textures[0].texture_view
        assert self._depth_texture.texture_view
        assert self._pick_texture.texture_view
        render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": self._render_textures[0].texture_view,
                    "resolve_target": None,
                    "load_value": (0, 0, 0, 0),  # LoadOp.load or color
                    "store_op": wgpu.StoreOp.store,
                },
                {
                    "view": self._pick_texture.texture_view,
                    "resolve_target": None,
                    "load_value": (0, 0, 0, 0),  # LoadOp.load or color
                    "store_op": wgpu.StoreOp.store,
                },
            ],
            depth_stencil_attachment={
                "view": self._depth_texture.texture_view,
                "depth_load_value": 1.0,  # depth is 0..1, make initial value as high as we can
                "depth_store_op": wgpu.StoreOp.store,
                "stencil_load_value": wgpu.LoadOp.load,
                "stencil_store_op": wgpu.StoreOp.store,
            },
            occlusion_query_set=None,
        )

        for wobject in q:
            wgpu_data = wobject._wgpu_pipeline_objects
            for pinfo in wgpu_data["render_pipelines"]:
                render_pass.set_pipeline(pinfo["pipeline"])
                for slot, vbuffer in pinfo["vertex_buffers"].items():
                    render_pass.set_vertex_buffer(
                        slot,
                        vbuffer._wgpu_buffer[1],
                        vbuffer.vertex_byte_range[0],
                        vbuffer.vertex_byte_range[1],
                    )
                for bind_group_id, bind_group in enumerate(pinfo["bind_groups"]):
                    render_pass.set_bind_group(bind_group_id, bind_group, [], 0, 99)
                # Draw with or without index buffer
                if pinfo["index_buffer"] is not None:
                    ibuffer = pinfo["index_buffer"]
                    render_pass.set_index_buffer(ibuffer, 0, ibuffer.size)
                    render_pass.draw_indexed(*pinfo["index_args"])
                else:
                    render_pass.draw(*pinfo["index_args"])

        render_pass.end_pass()

    def _update_stdinfo_buffer(self, camera, physical_size, logical_size):
        # Make sure we have a buffer object
        if not hasattr(self, "_wgpu_stdinfo_buffer"):
            self._wgpu_stdinfo_buffer = Buffer(
                array_from_shadertype(stdinfo_uniform_type), usage="uniform"
            )

        # Update its data
        stdinfo_data = self._wgpu_stdinfo_buffer.data
        stdinfo_data["cam_transform"].flat = camera.matrix_world_inverse.elements
        stdinfo_data["cam_transform_inv"].flat = camera.matrix_world.elements
        stdinfo_data["projection_transform"].flat = camera.projection_matrix.elements
        stdinfo_data[
            "projection_transform_inv"
        ].flat = camera.projection_matrix_inverse.elements
        # stdinfo_data["ndc_to_world"].flat = np.linalg.inv(stdinfo_data["cam_transform"] @ stdinfo_data["projection_transform"])
        stdinfo_data["physical_size"] = physical_size
        stdinfo_data["logical_size"] = logical_size
        # Upload to GPU
        self._wgpu_stdinfo_buffer.update_range(0, 1)
        self._update_buffer(self._wgpu_stdinfo_buffer)

    def get_render_list(self, scene: WorldObject, camera: Camera):
        """Given a scene object, get a flat list of objects to render."""

        # Collect items
        def visit(wobject):
            nonlocal q
            if wobject.visible and hasattr(wobject, "material"):
                q.append(wobject)

        q = []
        scene.traverse(visit)

        # Next, sort them from back-to-front
        def sort_func(wobject: WorldObject):
            z = (
                Vector3()
                .set_from_matrix_position(wobject.matrix_world)
                .apply_matrix4(proj_screen_matrix)
                .z
            )
            return wobject.render_order, z

        proj_screen_matrix = Matrix4().multiply_matrices(
            camera.projection_matrix, camera.matrix_world_inverse
        )
        q.sort(key=sort_func)
        return q

    def _ensure_up_to_date(self, wobject, force=False):
        """Update the GPU objects associated with the given wobject. Returns
        quickly if no changes are needed.
        """

        # Do we need to create the pipeline infos (from the renderfunc for this wobject)?
        if force or wobject.rev > getattr(wobject, "_wgpu_rev", 0):
            wobject._wgpu_rev = wobject.rev
            wobject._wgpu_pipeline_infos = self._create_pipeline_infos(wobject)
            wobject._wgpu_pipeline_res = self._collect_pipeline_resources(wobject)
            wobject._wgpu_pipeline_objects = None  # Invalidate

        # Early exit?
        if not wobject._wgpu_pipeline_infos:
            return

        # Check if we need to update any resources. The number of
        # resources should typically be small. We could implement a
        # hook in the resource's rev setter so we only have to check
        # one flag ... but let's not optimize prematurely.
        for kind, resource in wobject._wgpu_pipeline_res:
            our_version = getattr(resource, "_wgpu_" + kind, (-1, None))[0]
            if resource.rev > our_version:
                update_func = getattr(self, "_update_" + kind)
                update_func(resource)
                # one of self._update_buffer self._update_texture, self._update_texture_view, self._update_sampler

        # Create gpu objects?
        if wobject._wgpu_pipeline_objects is None:
            wobject._wgpu_pipeline_objects = self._create_pipeline_objects(wobject)

    def _create_pipeline_infos(self, wobject):
        """Use the render function for this wobject and material,
        and return a list of dicts representing pipelines in an abstract way.
        These dicts can then be turned into actual pipeline objects.
        """

        # Get render function for this world object,
        # and use it to get a high-level description of pipelines.
        renderfunc = registry.get_render_function(wobject)
        if renderfunc is None:
            raise ValueError(
                f"Could not get a render function for {wobject.__class__.__name__} "
                f"with {wobject.material.__class__.__name__}"
            )

        # Prepare info for the render function
        render_info = RenderInfo(
            stdinfo_uniform=self._wgpu_stdinfo_buffer,
        )

        # Call render function
        pipeline_infos = renderfunc(wobject, render_info)
        if not pipeline_infos:
            pipeline_infos = None
        else:
            assert isinstance(pipeline_infos, list)

        return pipeline_infos

    def _collect_pipeline_resources(self, wobject):

        pipeline_infos = wobject._wgpu_pipeline_infos or []

        pipeline_resources = []  # List, because order matters

        # Collect list of resources. That we can we can easily iterate over
        # dependent resource on each render call.
        for pipeline_info in pipeline_infos:
            assert isinstance(pipeline_info, dict)
            buffer = pipeline_info.get("index_buffer", None)
            if buffer is not None:
                pipeline_resources.append(("buffer", buffer))
            for buffer in pipeline_info.get("vertex_buffers", {}).values():
                pipeline_resources.append(("buffer", buffer))
            for key in pipeline_info.keys():
                if key.startswith("bindings"):
                    resources = pipeline_info[key]
                    if isinstance(resources, dict):
                        resources = resources.values()
                    for binding_type, resource in resources:
                        if binding_type.startswith("buffer/"):
                            assert isinstance(resource, Buffer)
                            pipeline_resources.append(("buffer", resource))
                        elif binding_type.startswith("sampler/"):
                            assert isinstance(resource, TextureView)
                            pipeline_resources.append(("sampler", resource))
                        elif binding_type.startswith("texture/"):
                            assert isinstance(resource, TextureView)
                            pipeline_resources.append(("texture", resource.texture))
                            pipeline_resources.append(("texture_view", resource))
                        elif binding_type.startswith("storage_texture/"):
                            assert isinstance(resource, TextureView)
                            pipeline_resources.append(("texture", resource.texture))
                            pipeline_resources.append(("texture_view", resource))
                        else:
                            raise RuntimeError(
                                f"Unknown resource binding type {binding_type}"
                            )

        return pipeline_resources

    def _create_pipeline_objects(self, wobject):
        """Generate wgpu pipeline objects from the list of pipeline info dicts."""

        # Prepare the three kinds of pipelines that we can get
        compute_pipelines = []
        render_pipelines = []
        alt_render_pipelines = []

        # Process each pipeline info object, converting each to a more concrete dict
        for pipeline_info in wobject._wgpu_pipeline_infos:
            if "vertex_shader" in pipeline_info and "fragment_shader" in pipeline_info:
                pipeline = self._compose_render_pipeline(wobject, pipeline_info)
                if pipeline_info.get("target", None) is None:
                    render_pipelines.append(pipeline)
                else:
                    raise NotImplementedError("Alternative render pipelines")
                    alt_render_pipelines.append(pipeline)
            elif "compute_shader" in pipeline_info:
                compute_pipelines.append(
                    self._compose_compute_pipeline(wobject, pipeline_info)
                )
            else:
                raise ValueError(
                    "Did not find compute_shader nor vertex_shader+fragment_shader in pipeline info."
                )

        return {
            "compute_pipelines": compute_pipelines,
            "render_pipelines": render_pipelines,
            "alt_render_pipelines": alt_render_pipelines,
        }

    def _compose_compute_pipeline(self, wobject, pipeline_info):
        """Given a high-level compute pipeline description, creates a
        lower-level representation that can be consumed by wgpu.
        """

        # todo: cache the pipeline with the shader (and entrypoint) as a hash

        device = self._device

        # Convert indices to args for the compute_pass.dispatch() call
        indices = pipeline_info["indices"]
        if not (
            isinstance(indices, tuple)
            and len(indices) == 3
            and all(isinstance(i, int) for i in indices)
        ):
            raise RuntimeError(
                f"Compute indices must be 3-tuple of ints, not {indices}."
            )
        index_args = indices

        # Get bind groups and pipeline layout from the buffers in pipeline_info.
        # This also makes sure the buffers and textures are up-to-date.
        bind_groups, pipeline_layout = self._get_bind_groups(pipeline_info)

        # Compile shader and create pipeline object
        cshader = pipeline_info["compute_shader"]
        cs_module = device.create_shader_module(code=cshader)
        compute_pipeline = device.create_compute_pipeline(
            layout=pipeline_layout,
            compute={"module": cs_module, "entry_point": "main"},
        )

        return {
            "pipeline": compute_pipeline,  # wgpu object
            "index_args": index_args,  # tuple
            "bind_groups": bind_groups,  # list of wgpu bind_group objects
        }

    def _compose_render_pipeline(self, wobject, pipeline_info):
        """Given a high-level render pipeline description, creates a
        lower-level representation that can be consumed by wgpu.
        """

        # todo: cache the pipeline with a lot of things as the hash
        # todo: cache vertex descriptors

        device = self._device

        # If an index buffer is present, update it, and get index_format.
        wgpu_index_buffer = None
        index_format = wgpu.IndexFormat.uint32
        index_buffer = pipeline_info.get("index_buffer", None)
        if index_buffer is not None:
            wgpu_index_buffer = index_buffer._wgpu_buffer[1]
            index_format = index_buffer.format

        # Convert and check high-level indices. Indices represent a range
        # of index id's, or define what indices in the index buffer are used.
        indices = pipeline_info.get("indices", None)
        if indices is None:
            if index_buffer is None:
                raise RuntimeError("Need indices or index_buffer ")
            indices = range(index_buffer.data.size)
        # Convert to 2-element tuple (vertex, instance)
        if not isinstance(indices, tuple):
            indices = (indices,)
        if len(indices) == 1:
            indices = indices + (1,)  # add instancing index
        if len(indices) != 2:
            raise RuntimeError("Render pipeline indices must be a 2-element tuple.")

        # Convert indices to args for the render_pass.draw() or draw_indexed()
        # draw(count_vertex, count_instance, first_vertex, first_instance)
        # draw_indexed(count_v, count_i, first_vertex, base_vertex, first_instance)
        index_args = [0, 0, 0, 0]
        for i, index in enumerate(indices):
            if isinstance(index, int):
                index_args[i] = index
            elif isinstance(index, range):
                assert index.step == 1
                index_args[i] = index.stop - index.start
                index_args[i + 2] = index.start
            else:
                raise RuntimeError(
                    "Render pipeline indices must be a 2-element tuple with ints or ranges."
                )
        if wgpu_index_buffer is not None:
            base_vertex = 0  # A value added to each index before reading [...]
            index_args.insert(-1, base_vertex)

        # Process vertex buffers. Update the buffer, and produces a descriptor.
        vertex_buffers = {}
        vertex_buffer_descriptors = []
        # todo: we can probably expose multiple attributes per buffer using a BufferView
        # -> can we also leverage numpy here?
        for slot, buffer in pipeline_info.get("vertex_buffers", {}).items():
            slot = int(slot)
            vbo_des = {
                "array_stride": buffer.nbytes // buffer.nitems,
                "step_mode": wgpu.InputStepMode.vertex,  # vertex or instance
                "attributes": [
                    {
                        "format": buffer.format,
                        "offset": 0,
                        "shader_location": slot,
                    }
                ],
            }
            vertex_buffers[slot] = buffer
            vertex_buffer_descriptors.append(vbo_des)

        # Get bind groups and pipeline layout from the buffers in pipeline_info.
        # This also makes sure the buffers and textures are up-to-date.
        bind_groups, pipeline_layout = self._get_bind_groups(pipeline_info)

        # Compile shaders
        vs_entry_point = fs_entry_point = "main"
        vshader = pipeline_info["vertex_shader"]
        if isinstance(vshader, tuple):
            vshader, vs_entry_point = vshader
        fshader = pipeline_info["fragment_shader"]
        if isinstance(fshader, tuple):
            fshader, fs_entry_point = fshader
        vs_module = self._get_shader_module(vshader, vshader)
        fs_module = self._get_shader_module(fshader, fshader)

        # Instantiate the pipeline object
        # todo: is this how strip_index_format is supposed to work?
        strip_index_format = 0
        if "strip" in pipeline_info["primitive_topology"]:
            strip_index_format = index_format
        pipeline = device.create_render_pipeline(
            layout=pipeline_layout,
            vertex={
                "module": vs_module,
                "entry_point": vs_entry_point,
                "buffers": vertex_buffer_descriptors,
            },
            primitive={
                "topology": pipeline_info["primitive_topology"],
                "strip_index_format": strip_index_format,
                "front_face": wgpu.FrontFace.ccw,
                "cull_mode": wgpu.CullMode.none,
            },
            depth_stencil={
                "format": self._depth_texture.format,
                "depth_write_enabled": True,  # optional
                "depth_compare": wgpu.CompareFunction.less,  # optional
                "stencil_front": {},  # use defaults
                "stencil_back": {},  # use defaults
                "depth_bias": 0,
                "depth_bias_slope_scale": 0.0,
                "depth_bias_clamp": 0.0,
            },
            multisample={
                "count": self._msaa,
                "mask": 0xFFFFFFFF,
                "alpha_to_coverage_enabled": False,
            },
            fragment={
                "module": fs_module,
                "entry_point": fs_entry_point,
                "targets": [
                    {
                        "format": self._render_textures[0].format,
                        "blend": {
                            "alpha": (
                                wgpu.BlendFactor.one,
                                wgpu.BlendFactor.zero,
                                wgpu.BlendOperation.add,
                            ),
                            "color": (
                                wgpu.BlendFactor.src_alpha,
                                wgpu.BlendFactor.one_minus_src_alpha,
                                wgpu.BlendOperation.add,
                            ),
                        },
                        "write_mask": wgpu.ColorWrite.ALL,
                    },
                    {
                        "format": self._pick_texture.format,
                        "blend": None,
                        "write_mask": wgpu.ColorWrite.ALL,
                    },
                ],
            },
        )

        return {
            "pipeline": pipeline,  # wgpu object
            "index_args": index_args,  # tuple
            "index_buffer": wgpu_index_buffer,  # Buffer
            "vertex_buffers": vertex_buffers,  # dict of slot -> Buffer
            "bind_groups": bind_groups,  # list of wgpu bind_group objects
        }

    def _get_bind_groups(self, pipeline_info):
        """Given high-level information on bindings, create the corresponding
        wgpu objects. This assumes that all buffers and textures are up-to-date.
        Returns (bind_groups, pipeline_layout).
        """
        # todo: cache bind_group_layout objects
        # todo: cache pipeline_layout objects
        # todo: can perhaps be more specific about visibility

        device = self._device

        # Collect resource groups (keys e.g. "bindings1", "bindings132")
        resource_groups = []
        for key in pipeline_info.keys():
            if key.startswith("bindings"):
                i = int(key[len("bindings") :])
                assert i >= 0
                if not pipeline_info[key]:
                    continue
                while len(resource_groups) <= i:
                    resource_groups.append({})
                resource_groups[i] = pipeline_info[key]

        # Create bind groups and bind group layouts
        bind_groups = []
        bind_group_layouts = []
        for resources in resource_groups:
            if not isinstance(resources, dict):
                resources = {slot: resource for slot, resource in enumerate(resources)}
            # Collect list of dicts
            bindings = []
            binding_layouts = []
            for slot, type_resource in resources.items():
                assert isinstance(type_resource, tuple) and len(type_resource) == 2
                binding_type, resource = type_resource
                subtype = binding_type.split("/")[-1]

                if binding_type.startswith("buffer/"):
                    assert isinstance(resource, Buffer)
                    bindings.append(
                        {
                            "binding": slot,
                            "resource": {
                                "buffer": resource._wgpu_buffer[1],
                                "offset": 0,
                                "size": resource.nbytes,
                            },
                        }
                    )
                    binding_layouts.append(
                        {
                            "binding": slot,
                            "visibility": visibility_all,
                            "buffer": {
                                "type": getattr(wgpu.BufferBindingType, subtype),
                                "has_dynamic_offset": False,
                                "min_binding_size": 0,
                            },
                        }
                    )
                elif binding_type.startswith("sampler/"):
                    assert isinstance(resource, TextureView)
                    bindings.append(
                        {"binding": slot, "resource": resource._wgpu_sampler[1]}
                    )
                    binding_layouts.append(
                        {
                            "binding": slot,
                            "visibility": wgpu.ShaderStage.FRAGMENT,
                            "sampler": {
                                "type": getattr(wgpu.SamplerBindingType, subtype),
                            },
                        }
                    )
                elif binding_type.startswith("texture/"):
                    assert isinstance(resource, TextureView)
                    bindings.append(
                        {"binding": slot, "resource": resource._wgpu_texture_view[1]}
                    )
                    dim = resource.view_dim
                    dim = getattr(wgpu.TextureViewDimension, dim, dim)
                    sample_type = getattr(wgpu.TextureSampleType, subtype, subtype)
                    # Derive sample type from texture
                    if sample_type == "auto":
                        fmt = ALTTEXFORMAT.get(resource.format, [resource.format])[0]
                        if "float" in fmt or "norm" in fmt:
                            sample_type = wgpu.TextureSampleType.float
                            # For float32 wgpu does not allow the sampler to be filterable,
                            # except when the native-only feature
                            # TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES is set,
                            # which wgpu-py does by default.
                            # if "32float" in fmt:
                            #     sample_type = wgpu.TextureSampleType.unfilterable_float
                        elif "uint" in fmt:
                            sample_type = wgpu.TextureSampleType.uint
                        elif "sint" in fmt:
                            sample_type = wgpu.TextureSampleType.sint
                        elif "depth" in fmt:
                            sample_type = wgpu.TextureSampleType.depth
                        else:
                            raise ValueError("Could not determine texture sample type.")
                    binding_layouts.append(
                        {
                            "binding": slot,
                            "visibility": wgpu.ShaderStage.FRAGMENT,
                            "texture": {
                                "sample_type": sample_type,
                                "view_dimension": dim,
                                "multisampled": False,
                            },
                        }
                    )
                elif binding_type.startswith("storage_texture/"):
                    assert isinstance(resource, TextureView)
                    bindings.append(
                        {"binding": slot, "resource": resource._wgpu_texture_view[1]}
                    )
                    dim = resource.view_dim
                    dim = getattr(wgpu.TextureViewDimension, dim, dim)
                    fmt = ALTTEXFORMAT.get(resource.format, [resource.format])[0]
                    binding_layouts.append(
                        {
                            "binding": slot,
                            "visibility": visibility_all,
                            "storage_texture": {
                                "access": getattr(wgpu.StorageTextureAccess, subtype),
                                "format": fmt,
                                "view_dimension": dim,
                            },
                        }
                    )

            # Create wgpu objects
            bind_group_layout = device.create_bind_group_layout(entries=binding_layouts)
            bind_group = device.create_bind_group(
                layout=bind_group_layout, entries=bindings
            )
            bind_groups.append(bind_group)
            bind_group_layouts.append(bind_group_layout)

        # Create pipeline layout object from list of layouts
        pipeline_layout = device.create_pipeline_layout(
            bind_group_layouts=bind_group_layouts
        )

        return bind_groups, pipeline_layout

    def _update_buffer(self, resource):
        buffer = getattr(resource, "_wgpu_buffer", (-1, None))[1]

        # todo: dispose an old buffer? / reuse an old buffer?

        pending_uploads = resource._pending_uploads
        resource._pending_uploads = []
        bytes_per_item = resource.nbytes // resource.nitems

        # Create buffer if needed
        if buffer is None or buffer.size != resource.nbytes:
            usage = wgpu.BufferUsage.COPY_DST
            for u in resource.usage.split("|"):
                usage |= getattr(wgpu.BufferUsage, u)
            buffer = self._device.create_buffer(size=resource.nbytes, usage=usage)

        queue = self._device.queue
        encoder = self._device.create_command_encoder()

        # Upload any pending data
        for offset, size in pending_uploads:
            subdata = resource._get_subdata(offset, size)
            # A: map the buffer, writes to it, then unmaps. But we don't offer a mapping API in wgpu-py
            # B: roll data in new buffer, copy from there to existing buffer
            tmp_buffer = self._device.create_buffer_with_data(
                data=subdata,
                usage=wgpu.BufferUsage.COPY_SRC,
            )
            boffset, bsize = bytes_per_item * offset, bytes_per_item * size
            encoder.copy_buffer_to_buffer(tmp_buffer, 0, buffer, boffset, bsize)
            # C: using queue. This may be sugar for B, but it may also be optimized
            # Unfortunately, this seems to crash the device :/
            # queue.write_buffer(buffer, bytes_per_item * offset, subdata, 0, subdata.nbytes)
            # D: A staging buffer/belt https://github.com/gfx-rs/wgpu-rs/blob/master/src/util/belt.rs
            # todo: look into staging buffers?

        queue.submit([encoder.finish()])
        resource._wgpu_buffer = resource.rev, buffer

    def _update_texture_view(self, resource):
        if resource._is_default_view:
            texture_view = resource.texture._wgpu_texture[1].create_view()
        else:
            dim = resource._view_dim
            assert resource._mip_range.step == 1
            assert resource._layer_range.step == 1
            texture_view = resource.texture._wgpu_texture[1].create_view(
                format=ALTTEXFORMAT.get(resource.format, [resource.format])[0],
                dimension=f"{dim}d" if isinstance(dim, int) else dim,
                aspect=resource._aspect,
                base_mip_level=resource._mip_range.start,
                mip_level_count=len(resource._mip_range),
                base_array_layer=resource._layer_range.start,
                array_layer_count=len(resource._layer_range),
            )
        resource._wgpu_texture_view = resource.rev, texture_view

    def _update_texture(self, resource):

        texture = getattr(resource, "_wgpu_texture", (-1, None))[1]
        pending_uploads = resource._pending_uploads
        resource._pending_uploads = []

        format = resource.format
        pixel_padding = None
        if format in ALTTEXFORMAT:
            format, pixel_padding, extra_bytes = ALTTEXFORMAT[format]

        # Create texture if needed
        if texture is None:  # todo: or needs to be replaced (e.g. resized)
            usage = wgpu.TextureUsage.COPY_DST
            for u in resource.usage.split("|"):
                usage |= getattr(wgpu.TextureUsage, u)
            texture = self._device.create_texture(
                size=resource.size,
                usage=usage,
                dimension=f"{resource.dim}d",
                format=getattr(wgpu.TextureFormat, format),
                mip_level_count=1,
                sample_count=1,  # msaa?
            )  # todo: let resource specify mip_level_count and sample_count

        bytes_per_pixel = resource.nbytes // (
            resource.size[0] * resource.size[1] * resource.size[2]
        )
        if pixel_padding is not None:
            bytes_per_pixel += extra_bytes

        queue = self._device.queue
        encoder = self._device.create_command_encoder()

        # Upload any pending data
        for offset, size in pending_uploads:
            subdata = resource._get_subdata(offset, size, pixel_padding)
            # B: using a temp buffer
            # tmp_buffer = self._device.create_buffer_with_data(data=subdata,
            #     usage=wgpu.BufferUsage.COPY_SRC,
            # )
            # encoder.copy_buffer_to_texture(
            #     {
            #         "buffer": tmp_buffer,
            #         "offset": 0,
            #         "bytes_per_row": size[0] * bytes_per_pixel,  # multiple of 256
            #         "rows_per_image": size[1],
            #     },
            #     {
            #         "texture": texture,
            #         "mip_level": 0,
            #         "origin": offset,
            #     },
            #     copy_size=size,
            # )
            # C: using the queue, which may be doing B, but may also be optimized,
            #    and the bytes_per_row limitation does not apply here
            queue.write_texture(
                {"texture": texture, "origin": offset, "mip_level": 0},
                subdata,
                {"bytes_per_row": size[0] * bytes_per_pixel, "rows_per_image": size[1]},
                size,
            )

        queue.submit([encoder.finish()])
        resource._wgpu_texture = resource.rev, texture

    def _update_sampler(self, resource):
        # A sampler's info (and raw object) are stored on a TextureView
        amodes = resource._address_mode.replace(",", " ").split() or ["clamp"]
        while len(amodes) < 3:
            amodes.append(amodes[-1])
        filters = resource._filter.replace(",", " ").split() or ["nearest"]
        while len(filters) < 3:
            filters.append(filters[-1])
        ammap = {"clamp": "clamp-to-edge", "mirror": "mirror-repeat"}
        sampler = self._device.create_sampler(
            address_mode_u=ammap.get(amodes[0], amodes[0]),
            address_mode_v=ammap.get(amodes[1], amodes[1]),
            address_mode_w=ammap.get(amodes[2], amodes[2]),
            mag_filter=filters[0],
            min_filter=filters[1],
            mipmap_filter=filters[2],
            # lod_min_clamp -> use default 0
            # lod_max_clamp -> use default inf
            # compare -> only not-None for comparison samplers!
        )
        resource._wgpu_sampler = resource.rev, sampler

    def _get_shader_module(self, key, source):
        """Compile a shader module object, or re-use it from the cache."""
        # todo: make this work for objects following the ShaderSourceTemplate interface
        # todo: also release shader modules that are no longer used
        if key not in self._shader_cache:
            m = self._device.create_shader_module(code=source)
            self._shader_cache[key] = m
        return self._shader_cache[key]

    # Picking

    def get_pick_info(self, pos):
        """Get information about the given window location. The given
        pos is a 2D point in logical pixels (with the origin at the
        top-left). Returns a dict with fields:

        * "ndc": The position in normalized device coordinates, the 3d element
            being the depth (0..1). Can be translated to the position
            in world coordinates using the camera transforms.
        * "rgba": The value in the color buffer. All zero's when rendering
          directly to the screen (bypassing post-processing).
        * "world_object": the object at that location (provided that
          the object supports picking).
        * Additional pick info may be available, depending on the type of
          object and its material. See the world-object classes for details.
        """

        # Make pos 0..1, so we can scale it to the render texture
        if self._logical_size:
            logical_size = self._logical_size
        else:
            logical_size = self._canvas.get_logical_size()
        float_pos = pos[0] / logical_size[0], pos[1] / logical_size[1]

        can_sample_color = self._render_textures[0].texture is not None

        # Sample
        encoder = self._device.create_command_encoder()
        self._copy_pixel(encoder, self._depth_texture, float_pos, 0)
        if can_sample_color:
            self._copy_pixel(encoder, self._render_textures[0], float_pos, 8)
        self._copy_pixel(encoder, self._pick_texture, float_pos, 16)
        queue = self._device.queue
        queue.submit([encoder.finish()])

        # Collect data from the buffer
        data = self._pixel_info_buffer.map_read()
        depth = data[0:4].cast("f")[0]
        color = tuple(data[8:12].cast("B"))
        pick_value = tuple(data[16:32].cast("i"))
        wobject = self._pick_map.get(pick_value[0], None)
        # Note: the position in world coordinates is not included because
        # it depends on the camera, but we don't "own" the camera.

        info = {
            "ndc": (2 * float_pos[0] - 1, 2 * float_pos[1] - 1, depth),
            "rgba": color if can_sample_color else (0, 0, 0, 0),
            "world_object": wobject,
        }

        if wobject and hasattr(wobject, "material"):
            pick_info = wobject.material._wgpu_get_pick_info(pick_value)
            info.update(pick_info)
        return info

    def _copy_pixel(self, encoder, render_texture, float_pos, buf_offset):

        # Map position to the texture index
        w, h, d = render_texture.size
        x = max(0, min(w - 1, int(float_pos[0] * w)))
        y = max(0, min(h - 1, int(float_pos[1] * h)))

        # Note: bytes_per_row must be a multiple of 256.
        encoder.copy_texture_to_buffer(
            {
                "texture": render_texture.texture,
                "mip_level": 0,
                "origin": (x, y, 0),
            },
            {
                "buffer": self._pixel_info_buffer,
                "offset": buf_offset,
                "bytes_per_row": 256,  # render_texture.bytes_per_pixel,
                "rows_per_image": 1,
            },
            copy_size=(1, 1, 1),
        )

    def snapshot(self, step_index=-1):
        """Create a snapshot of the currently rendered image."""

        # Prepare
        rt = self._render_textures[step_index]
        size = rt.size
        bytes_per_pixel = 4

        # Note, with queue.read_texture the bytes_per_row limitation does not apply.
        data = self._device.queue.read_texture(
            {
                "texture": rt.texture,
                "mip_level": 0,
                "origin": (0, 0, 0),
            },
            {
                "offset": 0,
                "bytes_per_row": bytes_per_pixel * size[0],
                "rows_per_image": size[1],
            },
            size,
        )

        return np.frombuffer(data, np.uint8).reshape(size[1], size[0], 4)



from math import acos, sin, cos, floor, ceil

from .utils import clamp
from .quaternion import Quaternion


class Vector3:
    def __init__(self, x: float = 0, y: float = 0, z: float = 0) -> None:
        self.x = x
        self.y = y
        self.z = z

    def __repr__(self) -> str:
        return f"Vector3({self.x}, {self.y}, {self.z})"

    def set(self, x: float, y: float, z: float) -> "Vector3":
        self.x = x
        self.y = y
        self.z = z
        return self

    def set_scalar(self, s: float) -> "Vector3":
        self.x = s
        self.y = s
        self.z = s
        return self

    def set_x(self, x: float) -> "Vector3":
        self.x = x
        return self

    def set_y(self, y: float) -> "Vector3":
        self.y = y
        return self

    def set_z(self, z: float) -> "Vector3":
        self.z = z
        return self

    def set_component(self, index: int, value: float) -> "Vector3":
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.z = value
        else:
            raise IndexError()
        return self

    def get_component(self, index: int) -> float:
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise IndexError()

    def clone(self) -> "Vector3":
        return Vector3(self.x, self.y, self.z)

    def copy(self, v: "Vector3") -> "Vector3":
        self.x = v.x
        self.y = v.y
        self.z = v.z
        return self

    def add(self, v: "Vector3") -> "Vector3":
        self.x += v.x
        self.y += v.y
        self.z += v.z
        return self

    def add_scalar(self, s: float) -> "Vector3":
        self.x += s
        self.y += s
        self.z += s
        return self

    def add_vectors(self, a: "Vector3", b: "Vector3") -> "Vector3":
        self.x = a.x + b.x
        self.y = a.y + b.y
        self.z = a.z + b.z
        return self

    def add_scaled_vector(self, v: "Vector3", s: float) -> "Vector3":
        self.x += v.x * s
        self.y += v.y * s
        self.z += v.z * s
        return self

    def sub(self, v: "Vector3") -> "Vector3":
        self.x -= v.x
        self.y -= v.y
        self.z -= v.z
        return self

    def sub_scalar(self, s: float) -> "Vector3":
        self.x -= s
        self.y -= s
        self.z -= s
        return self

    def sub_vectors(self, a: "Vector3", b: "Vector3") -> "Vector3":
        self.x = a.x - b.x
        self.y = a.y - b.y
        self.z = a.z - b.z
        return self

    def multiply(self, v: "Vector3") -> "Vector3":
        self.x *= v.x
        self.y *= v.y
        self.z *= v.z
        return self

    def multiply_scalar(self, s: float) -> "Vector3":
        self.x *= s
        self.y *= s
        self.z *= s
        return self

    def multiply_vectors(self, a: "Vector3", b: "Vector3") -> "Vector3":
        self.x = a.x * b.x
        self.y = a.y * b.y
        self.z = a.z * b.z
        return self

    def apply_euler(self, euler: "Euler") -> "Vector3":
        return self.apply_quaternion(_tmp_quaternion.set_from_euler(euler))

    def apply_axis_angle(self, axis: "Vector3", angle: float) -> "Vector3":
        return self.apply_quaternion(_tmp_quaternion.set_from_axis_angle(axis, angle))

    def apply_matrix3(self, m: "Matrix3") -> "Vector3":
        x, y, z = self.x, self.y, self.z
        e = m.elements

        self.x = e[0] * x + e[3] * y + e[6] * z
        self.y = e[1] * x + e[4] * y + e[7] * z
        self.z = e[2] * x + e[5] * y + e[8] * z
        return self

    def apply_normal_matrix(self, m: "Matrix3") -> "Vector3":
        return self.apply_matrix3(m).normalize()

    def apply_matrix4(self, m: "Matrix4") -> "Vector3":
        x, y, z = self.x, self.y, self.z
        e = m.elements

        denom = e[3] * x + e[7] * y + e[11] * z + e[15]
        if denom == 0:
            w = float("Inf")
        elif denom == -0:
            w = float("-Inf")
        else:
            w = 1 / denom
        self.x = (e[0] * x + e[4] * y + e[8] * z + e[12]) * w
        self.y = (e[1] * x + e[5] * y + e[9] * z + e[13]) * w
        self.z = (e[2] * x + e[6] * y + e[10] * z + e[14]) * w
        return self

    def apply_quaternion(self, q: "Quaternion") -> "Vector3":
        x = self.x
        y = self.y
        z = self.z
        qx = q.x
        qy = q.y
        qz = q.z
        qw = q.w

        # calculate quat * vector
        ix = qw * x + qy * z - qz * y
        iy = qw * y + qz * x - qx * z
        iz = qw * z + qx * y - qy * x
        iw = -qx * x - qy * y - qz * z

        # calculate result * inverse quat
        self.x = ix * qw + iw * -qx + iy * -qz - iz * -qy
        self.y = iy * qw + iw * -qy + iz * -qx - ix * -qz
        self.z = iz * qw + iw * -qz + ix * -qy - iy * -qx
        return self

    def project(self, camera) -> "Vector3":
        return self.apply_matrix4(camera.matrix_world_inverse).apply_matrix4(
            camera.projection_matrix
        )

    def unproject(self, camera) -> "Vector3":
        return self.apply_matrix4(camera.projection_matrix_inverse).apply_matrix4(
            camera.matrix_world
        )

    def transform_direction(self, m: "Matrix") -> "Vector3":
        # interpret self as directional vector
        # and apply affine transform in matrix4 m
        x = self.x
        y = self.y
        z = self.z
        e = m.elements

        self.x = e[0] * x + e[4] * y + e[8] * z
        self.y = e[1] * x + e[5] * y + e[9] * z
        self.z = e[2] * x + e[6] * y + e[10] * z
        return self.normalize()

    def divide(self, v: "Vector3") -> "Vector3":
        self.x /= v.x
        self.y /= v.y
        self.z /= v.z
        return self

    def divide_scalar(self, s: float) -> "Vector3":
        self.x /= s
        self.y /= s
        self.z /= s
        return self

    def min(self, v: "Vector3") -> "Vector3":
        self.x = min(self.x, v.x)
        self.y = min(self.y, v.y)
        self.z = min(self.z, v.z)
        return self

    def max(self, v: "Vector3") -> "Vector3":
        self.x = max(self.x, v.x)
        self.y = max(self.y, v.y)
        self.z = max(self.z, v.z)
        return self

    def clamp(self, min: "Vector3", max: "Vector3") -> "Vector3":
        # assumes min < max, component-wise
        self.x = clamp(self.x, min.x, max.x)
        self.y = clamp(self.y, min.y, max.y)
        self.z = clamp(self.z, min.z, max.z)
        return self

    def clamp_scalar(self, min: float, max: float) -> "Vector3":
        self.x = clamp(self.x, min, max)
        self.y = clamp(self.y, min, max)
        self.z = clamp(self.z, min, max)

    def clamp_length(self, min: float, max: float) -> "Vector3":
        length = self.length()
        return self.divide_scalar(length or 1).multiply_scalar(clamp(length, min, max))

    def floor(self) -> "Vector3":
        self.x = floor(self.x)
        self.y = floor(self.y)
        self.z = floor(self.z)
        return self

    def ceil(self) -> "Vector3":
        self.x = ceil(self.x)
        self.y = ceil(self.y)
        self.z = ceil(self.z)
        return self

    def round(self) -> "Vector3":
        self.x = round(self.x)
        self.y = round(self.y)
        self.z = round(self.z)
        return self

    def round_to_zero(self) -> "Vector3":
        self.x = ceil(self.x) if self.x < 0 else floor(self.x)
        self.y = ceil(self.y) if self.y < 0 else floor(self.y)
        self.z = ceil(self.z) if self.z < 0 else floor(self.z)
        return self

    def negate(self) -> "Vector3":
        self.x = -self.x
        self.y = -self.y
        self.z = -self.z
        return self

    def dot(self, v: "Vector3") -> float:
        return self.x * v.x + self.y * v.y + self.z * v.z

    def length_sq(self) -> float:
        return self.x ** 2 + self.y ** 2 + self.z ** 2

    def length(self) -> float:
        return self.length_sq() ** 0.5

    def manhattan_length(self) -> float:
        return abs(self.x) + abs(self.y) + abs(self.z)

    def normalize(self) -> "Vector3":
        return self.divide_scalar(self.length() or 1)

    def set_length(self, length: float) -> "Vector3":
        return self.normalize().multiply_scalar(length)

    def lerp(self, v: "Vector3", a: float) -> "Vector3":
        self.x += (v.x - self.x) * a
        self.y += (v.y - self.y) * a
        self.z += (v.z - self.z) * a
        return self

    def lerp_vectors(self, v1: "Vector3", v2: "Vector3", a: float) -> "Vector3":
        return self.sub_vectors(v2, v1).multiply_scalar(a).add(v1)

    def cross(self, v: "Vector3") -> "Vector3":
        return self.cross_vectors(self, v)

    def cross_vectors(self, a: "Vector3", b: "Vector3") -> "Vector3":
        ax = a.x
        ay = a.y
        az = a.z
        bx = b.x
        by = b.y
        bz = b.z

        self.x = ay * bz - az * by
        self.y = az * bx - ax * bz
        self.z = ax * by - ay * bx
        return self

    def project_on_vector(self, v: "Vector3") -> "Vector3":
        s = v.dot(self) / v.length_sq()
        return self.copy(v).multiply_scalar(s)

    def project_on_plane(self, n: "Vector3") -> "Vector3":
        _tmp_vector.copy(self).project_on_vector(n)
        return self.sub(_tmp_vector)

    def reflect(self, n: "Vector3") -> "Vector3":
        _tmp_vector.copy(n).multiply_scalar(2 * self.dot(n))
        return self.sub(_tmp_vector)

    def angle_to(self, v: "Vector3") -> float:
        denominator = (self.length_sq() * v.length_sq()) ** 0.5
        theta = self.dot(v) / denominator
        return acos(max(-1, min(1, theta)))

    def distance_to(self, v: "Vector3") -> float:
        return self.distance_to_squared(v) ** 0.5

    def distance_to_squared(self, v: "Vector3") -> float:
        dx = self.x - v.x
        dy = self.y - v.y
        dz = self.z - v.z

        return dx * dx + dy * dy + dz * dz

    def manhattan_distance_to(self, v: "Vector3") -> float:
        return abs(self.x - v.x) + abs(self.y - v.y) + abs(self.z - v.z)

    def set_from_spherical(self, s: "Spherical") -> "Vector3":
        return self.set_from_spherical_coords(s.radius, s.phi, s.theta)

    def set_from_spherical_coords(
        self, radius: float, phi: float, theta: float
    ) -> "Vector3":
        sin_phi_radius = sin(phi) * radius
        self.x = sin_phi_radius * sin(theta)
        self.y = cos(phi) * radius
        self.z = sin_phi_radius * cos(theta)
        return self

    def set_from_cylindrical(self, c: "Cylindrical") -> "Vector3":
        return self.set_from_cylindrical_coords(c.radius, c.theta, c.y)

    def set_from_cylindrical_coords(
        self, radius: float, theta: float, y: float
    ) -> "Vector3":
        self.x = radius * sin(theta)
        self.y = y
        self.z = radius * cos(theta)
        return self

    def set_from_matrix_position(self, m: "Matrix4") -> "Vector3":
        self.x = m.elements[12]
        self.y = m.elements[13]
        self.z = m.elements[14]
        return self

    def set_from_matrix_scale(self, m: "Matrix4") -> "Vector3":
        sx = self.set_from_matrix_column(m, 0).length()
        sy = self.set_from_matrix_column(m, 1).length()
        sz = self.set_from_matrix_column(m, 2).length()

        self.x = sx
        self.y = sy
        self.z = sz
        return self

    def set_from_matrix_column(self, m: "Matrix4", i: int) -> "Vector3":
        return self.from_array(m.elements, i * 4)

    def set_from_matrix3_column(self, m: "Matrix3", i: int) -> "Vector3":
        return self.from_array(m.elements, i * 3)

    def equals(self, v: "Vector3") -> bool:
        return self.x == v.x and self.y == v.y and self.z == v.z

    def __eq__(self, other: "Vector3") -> bool:
        return isinstance(other, Vector3) and self.equals(other)

    def from_array(self, array: list, offset: int = 0) -> "Vector3":
        self.x = array[offset]
        self.y = array[offset + 1]
        self.z = array[offset + 2]
        return self

    def to_array(self, array: list = None, offset: int = 0) -> list:
        if array is None:
            array = []

        padding = offset + 3 - len(array)
        if padding > 0:
            array.extend((None for _ in range(padding)))

        array[offset] = self.x
        array[offset + 1] = self.y
        array[offset + 2] = self.z
        return array

    def from_buffer_attribute(self, attribute, index: int) -> "Vector3":
        raise NotImplementedError()
        # self.x = attribute.getX(index)
        # self.y = attribute.getY(index)
        # self.z = attribute.getZ(index)
        # return self


_tmp_quaternion = Quaternion()
_tmp_vector = Vector3()



import numpy as np

from ..resources import Buffer
from ._base import Geometry


DTYPE = "f4"


def generate_torso(
    radius_bottom,
    radius_top,
    height,
    radial_segments,
    height_segments,
    theta_start,
    theta_length,
):
    # compute POSITIONS assuming x-y horizontal plane and z up axis

    # radius for each vertex ring from bottom to top
    n_rings = height_segments + 1
    radii = np.linspace(radius_bottom, radius_top, num=n_rings, dtype=DTYPE)

    # height for each vertex ring from bottom to top
    half_height = height / 2
    heights = np.linspace(-half_height, half_height, num=n_rings, dtype=DTYPE)

    # to enable texture mapping to fully wrap around the cylinder,
    # we can't close the geometry and need a degenerate vertex
    n_vertices = radial_segments + 1

    # xy coordinates on unit circle for a single vertex ring
    theta = np.linspace(
        theta_start, theta_start + theta_length, num=n_vertices, dtype=DTYPE
    )
    ring_xy = np.column_stack([np.cos(theta), np.sin(theta)])

    # put all the rings together
    positions = np.empty((n_rings, n_vertices, 3), dtype=DTYPE)
    positions[..., :2] = ring_xy[None, ...] * radii[:, None, None]
    positions[..., 2] = heights[:, None]

    # the NORMALS are the same for every ring, so compute for only one ring
    # and then repeat
    slope = (radius_bottom - radius_top) / height
    ring_normals = np.empty(positions.shape[1:], dtype=DTYPE)
    ring_normals[..., :2] = ring_xy
    ring_normals[..., 2] = slope
    ring_normals /= np.linalg.norm(ring_normals, axis=-1)[:, None]
    normals = np.empty_like(positions)
    normals[:] = ring_normals[None, ...]

    # the TEXTURE COORDS
    # u maps 0..1 to theta_start..theta_start+theta_length
    # v maps 0..1 to -height/2..height/2
    ring_u = (theta - theta_start) / theta_length
    ring_v = (heights / height) + 0.5
    texcoords = np.empty((n_rings, n_vertices, 2), dtype=DTYPE)
    texcoords[..., 0] = ring_u[None, :]
    texcoords[..., 1] = ring_v[:, None]

    # the face INDEX
    # the amount of vertices
    indices = np.arange(n_rings * n_vertices).reshape((n_rings, n_vertices))
    # for every panel (height_segments, radial_segments) there is a quad (2, 3)
    index = np.empty((height_segments, radial_segments, 2, 3), dtype="u4")
    # create a grid of initial indices for the panels
    index[:, :, 0, 0] = indices[
        np.arange(height_segments)[:, None], np.arange(radial_segments)[None, :]
    ]
    # the remainder of the indices for every panel are relative
    index[:, :, 0, 1] = index[:, :, 0, 0] + 1
    index[:, :, 0, 2] = index[:, :, 0, 0] + n_vertices
    index[:, :, 1, 0] = index[:, :, 0, 0] + n_vertices + 1
    index[:, :, 1, 1] = index[:, :, 1, 0] - 1
    index[:, :, 1, 2] = index[:, :, 1, 0] - n_vertices

    return (
        positions.reshape((-1, 3)),
        normals.reshape((-1, 3)),
        texcoords.reshape((-1, 2)),
        index.flatten(),
    )


def generate_cap(radius, height, radial_segments, theta_start, theta_length, up=True):
    # compute POSITIONS assuming x-y horizontal plane and z up axis

    # to enable texture mapping to fully wrap around the cylinder,
    # we can't close the geometry and need a degenerate vertex
    n_vertices = radial_segments + 1

    # xy coordinates on unit circle for vertex ring
    theta = np.linspace(
        theta_start, theta_start + theta_length, num=n_vertices, dtype=DTYPE
    )
    ring_xy = np.column_stack([np.cos(theta), np.sin(theta)])

    # put the vertices together, inserting a center vertex at the start
    positions = np.empty((1 + n_vertices, 3), dtype=DTYPE)
    positions[0, :2] = [0.0, 0.0]
    positions[1:, :2] = ring_xy * radius
    positions[..., 2] = height

    # the NORMALS
    normals = np.zeros_like(positions, dtype=DTYPE)
    sign = int(up) * 2.0 - 1.0
    normals[..., 2] = sign

    # the TEXTURE COORDS
    # uv etches out a circle from the [0..1, 0..1] range
    # direction is reversed for up=False
    texcoords = np.empty((1 + n_vertices, 2), dtype=DTYPE)
    texcoords[0] = [0.5, 0.5]
    texcoords[1:, 0] = ring_xy[:, 0] * 0.5 + 0.5
    texcoords[1:, 1] = ring_xy[:, 1] * 0.5 * sign + 0.5

    # the face INDEX
    indices = np.arange(n_vertices) + 1
    # for every radial segment there is a triangle (3)
    index = np.empty((radial_segments, 3), dtype="u4")
    # create a grid of initial indices for the panels
    index[:, 0] = indices[np.arange(radial_segments)]
    # the remainder of the indices for every panel are relative
    index[:, 1 + int(up)] = n_vertices
    index[:, 2 - int(up)] = index[:, 0] + 1

    return (
        positions,
        normals,
        texcoords,
        index.flatten(),
    )


def merge(groups):
    positions = np.concatenate([g[0] for g in groups])
    normals = np.concatenate([g[1] for g in groups])
    texcoords = np.concatenate([g[2] for g in groups])
    index = np.concatenate([g[3] for g in groups])
    i = 0
    j = 0
    for g in groups:
        index[i:] += j
        # advance cursor to start of next group index
        i += len(g[3])
        j = len(g[0])
    return positions, normals, texcoords, index


class CylinderGeometry(Geometry):
    def __init__(
        self,
        radius_bottom=1.0,
        radius_top=1.0,
        height=1.0,
        radial_segments=8,
        height_segments=1,
        theta_start=0.0,
        theta_length=np.pi * 2,
        open_ended=False,
    ):
        super().__init__()

        assert radial_segments > 0
        assert height_segments > 0
        assert theta_length != 0.0

        mesh = generate_torso(
            radius_bottom,
            radius_top,
            height,
            radial_segments,
            height_segments,
            theta_start,
            theta_length,
        )

        if not open_ended:
            groups = [mesh]
            if radius_bottom > 0:
                bottom_cap = generate_cap(
                    radius_bottom,
                    -height / 2,
                    radial_segments,
                    theta_start,
                    theta_length,
                    up=False,
                )
                groups.append(bottom_cap)
            if radius_top > 0:
                top_cap = generate_cap(
                    radius_top,
                    height / 2,
                    radial_segments,
                    theta_start,
                    theta_length,
                    up=True,
                )
                groups.append(top_cap)
            mesh = merge(groups)

        positions, normals, texcoords, index = mesh
        self.positions = Buffer(positions, usage="vertex|storage")
        self.normals = Buffer(normals, usage="vertex|storage")
        self.texcoords = Buffer(texcoords, usage="vertex|storage")
        self.index = Buffer(index, usage="index|storage")



# flake8: noqa

from ._wgpurenderer import stdinfo_uniform_type
from ._wgpurenderer import WgpuRenderer, registry, register_wgpu_render_function
from .postprocessing import PostProcessingStep
from . import meshrender
from . import linerender
from . import pointsrender
from . import volumerender
from . import backgroundrender



import numpy as np


def array_from_shadertype(shadertype):
    """Get a numpy array object from a dict shadertype."""
    assert isinstance(shadertype, dict)

    # Unravel the dict
    dtype_fields = []
    for name, type_tuple in shadertype.items():
        dtype_fields.append((name,) + tuple(type_tuple))

    # Create a scalar of this type
    uniform_data = np.zeros((), dtype=dtype_fields)
    return uniform_data


def normals_from_vertices(rr, tris):
    """Efficiently compute vertex normals for a triangulated surface."""
    # This code was taken from Vispy
    # ensure highest precision for our summation/vectorization "trick"
    rr = rr[:, :3].astype(np.float64)
    tris = tris.reshape(-1, 3)
    # first, compute triangle normals
    r1 = rr[tris[:, 0], :]
    r2 = rr[tris[:, 1], :]
    r3 = rr[tris[:, 2], :]
    tri_nn = np.cross((r2 - r1), (r3 - r1))

    # Triangle normals and areas
    size = np.sqrt(np.sum(tri_nn * tri_nn, axis=1))
    size[size == 0] = 1.0  # prevent ugly divide-by-zero
    tri_nn /= size[:, np.newaxis]

    npts = len(rr)

    # the following code replaces this, but is faster (vectorized):
    #
    # for p, verts in enumerate(tris):
    #     nn[verts, :] += tri_nn[p, :]
    #
    nn = np.zeros((npts, 3))
    for verts in tris.T:  # note this only loops 3x (number of verts per tri)
        for idx in range(3):  # x, y, z
            nn[:, idx] += np.bincount(
                verts.astype(np.int32), tri_nn[:, idx], minlength=npts
            )
    size = np.sqrt(np.sum(nn * nn, axis=1))
    size[size == 0] = 1.0  # prevent ugly divide-by-zero
    nn /= size[:, np.newaxis]
    return nn.astype(np.float32)



from enum import Enum
from math import asin, atan2

from .utils import clamp
from .matrix4 import Matrix4
from .quaternion import Quaternion


class Euler:
    class RotationOrders(Enum):
        XYZ = "XYZ"
        YZX = "YZX"
        ZXY = "ZXY"
        XZY = "XZY"
        YXZ = "YXZ"
        ZYX = "ZYX"

    DefaultOrder = RotationOrders.XYZ

    def __init__(
        self, x: float = 0, y: float = 0, z: float = 0, order: RotationOrders = None
    ) -> None:
        self.x = x
        self.y = y
        self.z = z
        self.order = order if order is not None else self.DefaultOrder

    def __repr__(self) -> str:
        return f"Euler({self.x}, {self.y}, {self.z}, {self.order})"

    def set(
        self, x: float, y: float, z: float, order: RotationOrders = None
    ) -> "Euler":
        self.x = x
        self.y = y
        self.z = z
        self.order = order if order is not None else self.order
        return self

    def clone(self) -> "Euler":
        return Euler(self.x, self.y, self.z, self.order)

    def copy(self, euler: "Euler") -> "Euler":
        self.x = euler.x
        self.y = euler.y
        self.z = euler.z
        self.order = euler.order
        return self

    def set_from_rotation_matrix(
        self, m: "Matrix4", order: RotationOrders = None
    ) -> "Euler":
        # assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)
        te = m.elements
        m11 = te[0]
        m12 = te[4]
        m13 = te[8]
        m21 = te[1]
        m22 = te[5]
        m23 = te[9]
        m31 = te[2]
        m32 = te[6]
        m33 = te[10]

        order = order or self.order

        if order == Euler.RotationOrders.XYZ:
            self.y = asin(clamp(m13, -1, 1))
            if abs(m13) < 0.9999999:
                self.x = atan2(-m23, m33)
                self.z = atan2(-m12, m11)
            else:
                self.x = atan2(m32, m22)
                self.z = 0

        elif order == Euler.RotationOrders.YXZ:
            self.x = asin(-clamp(m23, -1, 1))
            if abs(m23) < 0.9999999:
                self.y = atan2(m13, m33)
                self.z = atan2(m21, m22)
            else:
                self.y = atan2(-m31, m11)
                self.z = 0

        elif order == Euler.RotationOrders.ZXY:
            self.x = asin(clamp(m32, -1, 1))
            if abs(m32) < 0.9999999:
                self.y = atan2(-m31, m33)
                self.z = atan2(-m12, m22)
            else:
                self.y = 0
                self.z = atan2(m21, m11)

        elif order == Euler.RotationOrders.ZYX:
            self.y = asin(-clamp(m31, -1, 1))
            if abs(m31) < 0.9999999:
                self.x = atan2(m32, m33)
                self.z = atan2(m21, m11)
            else:
                self.x = 0
                self.z = atan2(-m12, m22)

        elif order == Euler.RotationOrders.YZX:
            self.z = asin(clamp(m21, -1, 1))
            if abs(m21) < 0.9999999:
                self.x = atan2(-m23, m22)
                self.y = atan2(-m31, m11)
            else:
                self.x = 0
                self.y = atan2(m13, m33)

        elif order == Euler.RotationOrders.XZY:
            self.z = asin(-clamp(m12, -1, 1))
            if abs(m12) < 0.9999999:
                self.x = atan2(m32, m22)
                self.y = atan2(m13, m11)
            else:
                self.x = atan2(-m23, m33)
                self.y = 0

        else:
            raise ValueError(f"{order} not supported")

        self.order = order

        return self

    def set_from_quaternion(
        self, q: "Quaternion", order: RotationOrders = None
    ) -> "Euler":
        _tmp_matrix4.make_rotation_from_quaternion(q)
        return self.set_from_rotation_matrix(_tmp_matrix4, order=order)

    def set_from_vector3(self, v: "Vector3", order: RotationOrders = None) -> "Euler":
        return self.set(v.x, v.y, v.z, order=order)

    def reorder(self, new_order: RotationOrders) -> "Euler":
        # warning: revolution info lost
        _tmp_quaternion.set_from_euler(self)
        return self.set_from_quaternion(_tmp_quaternion, order=new_order)

    def equals(self, euler: "Euler") -> bool:
        return (
            euler.x == self.x
            and euler.y == self.y
            and euler.z == self.z
            and euler.order == self.order
        )

    def __eq__(self, other: "Euler") -> bool:
        return isinstance(other, Euler) and self.equals(other)

    def from_array(self, array: list) -> "Euler":
        self.x = array[0]
        self.y = array[1]
        self.z = array[2]
        if len(array) >= 4:
            self.order = array[3]
        return self

    def to_array(self, array: list = None, offset: int = 0) -> list:
        if array is None:
            array = []

        padding = offset + 4 - len(array)
        if padding > 0:
            array.extend((None for _ in range(padding)))

        array[offset] = self.x
        array[offset + 1] = self.y
        array[offset + 2] = self.z
        array[offset + 3] = self.order
        return array

    def to_vector3(self, output: "Vector3" = None) -> "Vector3":
        from .vector3 import Vector3

        if output is not None:
            return output.set(self.x, self.y, self.z)
        else:
            return Vector3(self.x, self.y, self.z)


_tmp_matrix4 = Matrix4()
_tmp_quaternion = Quaternion()



import numpy as np

import wgpu

STRUCT_FORMAT_ALIASES = {"c": "B", "l": "i", "L": "I"}


class Resource:
    pass


class Buffer(Resource):
    """A buffer object represents a piece of memory to the GPU, that can be
    used as index buffer, vertex buffer, uniform buffer, or storage buffer.
    You can provide (and update data for it), or use it as a placeholder
    for a buffer with no representation on the CPU.

    Parameters:
        data (array, optional): Array data of any type that supports the
            buffer-protocol, (e.g. a bytes or numpy array). If not given
            or None, nbytes and nitems must be provided. The data is
            copied if it's float64 or not contiguous.
        usage (str): The way(s) that the texture will be used. E.g. "INDEX"
            "VERTEX", "UNIFORM". Multiple values can be given separated
            with "|". See wgpu.BufferUsage.
        nbytes (int): The number of bytes. If data is given, it is derived.
        nitems (int): The number of items. If data is given, it is derived.
        format (str): The format to use when used as a vertex buffer.
            Must be a value from wgpu.VertexFormat. By default it is
            derived from the data. Set when data is not given or when
            you want to overload the derived value.
    """

    def __init__(
        self,
        data=None,
        *,
        usage,
        nbytes=None,
        nitems=None,
        format=None,
    ):
        self._rev = 0
        # To specify the buffer size
        self._nbytes = 0
        self._nitems = 1
        self._format = format
        # The actual data (optional)
        self._data = None
        self._pending_uploads = []  # list of (offset, size) tuples

        # Get nbytes
        if data is not None:
            mem = memoryview(data)
            if mem.format == "d":
                raise ValueError("Float64 data is not supported, use float32 instead.")
            # if not mem.contiguous or mem.format == "d":
            #     format = "f" if mmemformat == "d" else mem.format
            #     x = np.empty(mem.shape, format)
            #     x[:] = mem
            #     mem = memoryview(x)
            self._data = data
            self._mem = mem
            self._nbytes = mem.nbytes
            self._nitems = mem.shape[0] if mem.shape else 1
            self._pending_uploads.append((0, self._nitems))
            if nbytes is not None and nbytes != self._nbytes:
                raise ValueError("Given nbytes does not match size of given data.")
            if nitems is not None and nitems != self._nitems:
                raise ValueError("Given nitems does not match shape of given data.")
        elif nbytes is not None and nitems is not None:
            self._nbytes = int(nbytes)
            self._nitems = int(nitems)
        else:
            raise ValueError(
                "Buffer must be instantiated with either data or nbytes and nitems."
            )

        # Determine usage
        if isinstance(usage, str):
            usages = usage.upper().replace(",", " ").replace("|", " ").split()
            assert usages
            self._usage = "|".join(usages)
        else:
            raise TypeError("Buffer usage must be str.")

        # We can use a subset when used as a vertex buffer
        self._vertex_byte_range = (0, self._nbytes)

    @property
    def rev(self):
        """An integer that is increased when update_range() is called."""
        return self._rev

    @property
    def usage(self):
        """The buffer usage flags (as a string with "|" as a separator)."""
        return self._usage

    @property
    def data(self):
        """The data for this buffer. Can be None if the data only
        exists on the GPU.

        Note: the data is the same reference that was given to instantiate this object,
        but this may change.
        """
        return self._data

    @property
    def mem(self):
        """The data for this buffer as a memoryview. Can be None if
        the data only exists on the GPU.
        """
        return self._mem

    @property
    def nbytes(self):
        """The number of bytes in the buffer."""
        return self._nbytes

    @property
    def nitems(self):
        """The number of items in the buffer."""
        return self._nitems

    @property
    def format(self):
        """The vertex or index format (depending on the value of usage)."""
        if self._format is not None:
            return self._format
        elif self.data is not None:
            self._format = format_from_memoryview(self.mem, self.usage)
            return self._format
        else:
            raise ValueError("Buffer has no data nor format.")

    @property
    def vertex_byte_range(self):
        """The offset and size, in bytes, when used as a vertex buffer."""
        return self._vertex_byte_range

    @vertex_byte_range.setter
    def vertex_byte_range(self, offset_nbytes):
        offset, nbytes = int(offset_nbytes[0]), int(offset_nbytes[1])
        assert offset >= 0
        assert offset + nbytes <= self.nbytes
        self._vertex_byte_range = offset, nbytes

    def update_range(self, offset=0, size=2 ** 50):
        """Mark a certain range of the data for upload to the GPU. The
        offset and size are expressed in integer number of elements.
        """
        # See ThreeJS BufferAttribute.updateRange
        # Check input
        assert isinstance(offset, int) and isinstance(size, int)
        if size == 0:
            return
        elif size < 0:
            raise ValueError("Update size must not be negative")
        elif offset < 0:
            raise ValueError("Update offset must not be negative")
        elif offset + size > self.nitems:
            size = self.nitems - offset
        # Merge with current entry?
        if self._pending_uploads:
            cur_offset, cur_size = self._pending_uploads.pop(-1)
            offset = min(offset, cur_offset)
            size = max(size, cur_size)
        # Limit and apply
        self._pending_uploads.append((offset, size))
        self._rev += 1
        # todo: this can be smarter, we have logic for chunking in the morph tool

    def _get_subdata(self, offset, size):
        """Return subdata as a contiguous array."""
        # If this is a full range, this is easy
        if offset == 0 and size == self.nitems and self.mem.contiguous:
            return self.mem
        # Get a numpy array, because memoryviews do not support nd slicing
        if isinstance(self.data, np.ndarray):
            arr = self.data
        elif not self.mem.c_contiguous:
            raise ValueError(
                "Non-contiguous texture data is only supported for numpy array."
            )
        else:
            arr = np.frombuffer(self.mem, self.mem.format).reshape(self.mem.shape)
        # Slice it
        sub_arr = arr[offset : offset + size]
        return memoryview(np.ascontiguousarray(sub_arr))


def format_from_memoryview(mem, usage):

    if "INDEX" in usage:

        format = str(mem.format)
        format = STRUCT_FORMAT_ALIASES.get(format, format)
        mapping = {
            "h": wgpu.IndexFormat.uint16,
            "H": wgpu.IndexFormat.uint16,
            "i": wgpu.IndexFormat.uint32,
            "I": wgpu.IndexFormat.uint32,
        }
        try:
            return mapping[format]
        except KeyError:
            raise TypeError(
                f"Need 16bit or 32bit signed/unsigned int (hHiI) for index data, not '{format}'."
            )

    else:  # if "VERTEX" in self.usage:

        shape = mem.shape
        if len(shape) == 1:
            shape = shape + (1,)
        assert len(shape) == 2
        format = str(mem.format)
        format = STRUCT_FORMAT_ALIASES.get(format, format)
        key = format, shape[-1]
        mapping = {
            ("f", 1): wgpu.VertexFormat.float32,
            ("f", 2): wgpu.VertexFormat.float32x2,
            ("f", 3): wgpu.VertexFormat.float32x3,
            ("f", 4): wgpu.VertexFormat.float32x4,
            #
            ("e", 2): wgpu.VertexFormat.float16x2,
            ("e", 4): wgpu.VertexFormat.float16x4,
            #
            ("b", 2): wgpu.VertexFormat.sint8x2,
            ("b", 4): wgpu.VertexFormat.sint8x4,
            ("B", 2): wgpu.VertexFormat.uint8x2,
            ("B", 4): wgpu.VertexFormat.uint8x4,
            #
            ("h", 2): wgpu.VertexFormat.sint16x2,
            ("h", 4): wgpu.VertexFormat.sint16x4,
            ("H", 2): wgpu.VertexFormat.uint16x2,
            ("H", 4): wgpu.VertexFormat.uint16x4,
            #
            ("i", 1): wgpu.VertexFormat.sint32,
            ("i", 2): wgpu.VertexFormat.sint32x2,
            ("i", 3): wgpu.VertexFormat.sint32x3,
            ("i", 4): wgpu.VertexFormat.sint32x4,
            #
            ("I", 1): wgpu.VertexFormat.uint32,
            ("I", 2): wgpu.VertexFormat.uint32x2,
            ("I", 3): wgpu.VertexFormat.uint32x3,
            ("I", 4): wgpu.VertexFormat.uint32x4,
        }
        try:
            return mapping[key]
        except KeyError:
            if format in ("d", "float64"):
                raise ValueError(
                    "64-bit float is not supported, use 32-bit float instead"
                )
            raise ValueError(f"Invalid format/shape for vertex data: {key}")



from ._base import Material
from ..utils import array_from_shadertype
from ..resources import Buffer


class PointsMaterial(Material):
    """The default material used by Points. Renders (antialiased) disks
    of the given size and color.
    """

    uniform_type = dict(color=("float32", 4), size=("float32",))

    def __init__(self, **kwargs):
        super().__init__()

        self.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), usage="UNIFORM"
        )

        self._map = None
        self.color = 1, 1, 1, 1
        self.size = 1

        for argname, val in kwargs.items():
            if not hasattr(self, argname):
                raise AttributeError(f"No attribute '{argname}'")
            setattr(self, argname, val)

    def _wgpu_get_pick_info(self, pick_value):
        # The instance is zero while renderer doesn't support instancing
        instance = pick_value[1]
        vertex = pick_value[2]
        return {"instance_index": instance, "vertex_index": vertex}

    @property
    def color(self):
        """The color of the points (if map is not set)."""
        return self.uniform_buffer.data["color"]

    @color.setter
    def color(self, color):
        self.uniform_buffer.data["color"] = tuple(color)
        self.uniform_buffer.update_range(0, 1)

    # @property
    # def map(self):
    #     """ The 1D texture map specifying the color for each point.
    #     """
    #     return self._map
    #
    # @map.setter
    # def map(self, map):
    #     self._map = map

    @property
    def size(self):
        """The size (diameter) of the points, in logical pixels."""
        return self.uniform_buffer.data["size"]

    @size.setter
    def size(self, size):
        self.uniform_buffer.data["size"] = size

    # todo: sizeAttenuation


class GaussianPointsMaterial(PointsMaterial):
    """A material for points, renders Gaussian blobs with a standard
    deviation of 1/6 of the size.
    """


# idea: a MarkerMaterial with more options for the shape, and an edge around the shape.
# Though perhaps such a material should be part of a higher level plotting lib.



from ..linalg import Matrix4
from ..objects._base import WorldObject


class Camera(WorldObject):
    """Abstract base camera. The purpose of a camera is to project
    world coordinates to normalized device coordinates (NDC). This is
    done by the (inverse of) the camera's own world matrix and the
    camera's projection transform. The former represent the camera's
    position, the latter is specific to the type of camera.

    Note that we follow the NDC coordinate system of WGPU, where
    x and y are in the range 0..1, z is in the range 0..1, and (-1, -1, 0)
    represents the bottom left corner.

    """

    def __init__(self):
        super().__init__()

        self.matrix_world_inverse = Matrix4()
        self.projection_matrix = Matrix4()
        self.projection_matrix_inverse = Matrix4()

    def set_viewport_size(self, width, height):
        # In logical pixels
        pass

    def update_matrix_world(self, *args, **kwargs):
        super().update_matrix_world(*args, **kwargs)
        self.matrix_world_inverse.get_inverse(self.matrix_world)

    def update_projection_matrix(self):
        raise NotImplementedError()


class NDCCamera(Camera):
    """A Camera operating in NDC coordinates: its projection matrix
    is the identity transform (but its matrix_world can still be set).

    In the NDC coordinate system of WGPU (and pygfx), x and y are in
    the range 0..1, z is in the range 0..1, and (-1, -1, 0) represents
    the bottom left corner."""

    def update_projection_matrix(self):
        eye = 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1
        self.projection_matrix.set(*eye)
        self.projection_matrix_inverse.set(*eye)


class ScreenCoordsCamera(Camera):
    """A Camera operating in screen coordinates. The depth range is the same
    as in NDC (0 to 1).
    """

    def __init__(self):
        super().__init__()
        self._width = 1
        self._height = 1

    def set_viewport_size(self, width, height):
        self._width = width
        self._height = height

    def update_projection_matrix(self):
        sx, sy, sz = 2 / self._width, 2 / self._height, 1
        dx, dy, dz = -1, -1, 0
        m = sx, 0, 0, dx, 0, sy, 0, dy, 0, 0, sz, dz, 0, 0, 0, 1
        self.projection_matrix.set(*m)
        self.projection_matrix_inverse.get_inverse(self.projection_matrix)



import numpy as np

from .. import Geometry, Line, LineSegmentMaterial


class AxesHelper(Line):
    def __init__(self, length=1.0, thickness=6.0):
        positions = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 0, 0],
                [0, 1, 0],
                [0, 0, 0],
                [0, 0, 1],
            ],
            dtype="f4",
        )
        positions *= length

        colors = np.array(
            [
                [1, 0.6, 0, 1],
                [1, 0.6, 0, 1],  # x is orange-ish
                [0.6, 1, 0, 1],
                [0.6, 1, 0, 1],  # y is yellow-ish
                [0, 0.6, 1, 1],
                [0, 0.6, 1, 1],  # z is blue-ish
            ],
            dtype="f4",
        )

        geometry = Geometry(positions=positions, colors=colors)
        material = LineSegmentMaterial(thickness=thickness, vertex_colors=True)

        super().__init__(geometry, material)



# flake8: noqa

from ._base import Material
from ._mesh import (
    MeshBasicMaterial,
    MeshNormalMaterial,
    MeshNormalLinesMaterial,
    MeshPhongMaterial,
    MeshSliceMaterial,
)
from ._points import PointsMaterial, GaussianPointsMaterial
from ._line import (
    LineMaterial,
    LineThinMaterial,
    LineThinSegmentMaterial,
    LineSegmentMaterial,
    LineArrowMaterial,
)
from ._volume import VolumeBasicMaterial, VolumeSliceMaterial
from ._background import BackgroundMaterial, BackgroundImageMaterial



from ..utils import array_from_shadertype
from ..resources import Buffer
from ._base import Material


class VolumeBasicMaterial(Material):
    """Base volume material."""

    uniform_type = dict(
        clim=("float32", 2),
    )

    def __init__(self, **kwargs):
        super().__init__()

        self.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), usage="UNIFORM"
        )

        self._map = None
        self.clim = 0, 1

        for argname, val in kwargs.items():
            if not hasattr(self, argname):
                raise AttributeError(f"No attribute '{argname}'")
            setattr(self, argname, val)

    def _wgpu_get_pick_info(self, pick_value):
        size = self.map.size
        x, y, z = [(v / 1048576) * s - 0.5 for v, s in zip(pick_value[1:], size)]
        return {"instance_index": 0, "voxel_index": (x, y, z)}

    @property
    def map(self):
        """The 3D texture representing the volume."""
        return self._map

    @map.setter
    def map(self, map):
        self._map = map

    @property
    def clim(self):
        """The contrast limits to apply to the map. Default (0, 1)"""
        return self.uniform_buffer.data["clim"]

    @clim.setter
    def clim(self, clim):
        self.uniform_buffer.data["clim"] = clim
        self.uniform_buffer.update_range(0, 1)


class VolumeSliceMaterial(VolumeBasicMaterial):
    """A material for rendering a slice through a 3D texture at the surface of a mesh.
    This material is not affected by lights.
    """

    uniform_type = dict(
        plane=("float32", 4),
        clim=("float32", 2),
    )

    @property
    def plane(self):
        """The plane to slice at, represented with 4 floats ``(a, b, c, d)``,
        which make up the equation: ``ax + by + cz + d = 0`` The plane
        definition applies to the world space (of the scene).
        """
        return self.uniform_buffer.data["plane"]

    @plane.setter
    def plane(self, plane):
        self.uniform_buffer.data["plane"] = plane
        self.uniform_buffer.update_range(0, 1)



from ._base import Material
from ..utils import array_from_shadertype
from ..resources import Buffer

# todo: in ThreeJS you can simply set a CubeTexture as the scene.background
# we could do that, and the scene could use these objects automatically.
# Or just leave it like this. I kinda like it.


class BackgroundMaterial(Material):
    """A background material that draws the background is a uniform color
    or in a gradient.
    """

    uniform_type = dict(
        color_bottom_left=("float32", 4),
        color_bottom_right=("float32", 4),
        color_top_left=("float32", 4),
        color_top_right=("float32", 4),
    )

    def __init__(self, *colors):
        super().__init__()

        self.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), usage="UNIFORM"
        )

        self.set_color(*colors)

    def set_color(self, *colors):
        """Set the background color. If one color is given, it will be used
        as a uniform color. If two colors are given, it will be used for
        the botton and top. If four colors are given, it will be used for the
        four corners.
        """
        if len(colors) == 0:
            self.color_bottom_left = (0, 0, 0, 1)
            self.color_bottom_right = (0, 0, 0, 1)
            self.color_top_left = (0, 0, 0, 1)
            self.color_top_right = (0, 0, 0, 1)
        elif len(colors) == 1:
            self.color_bottom_left = colors[0]
            self.color_bottom_right = colors[0]
            self.color_top_left = colors[0]
            self.color_top_right = colors[0]
        elif len(colors) == 2:
            self.color_bottom_left = colors[0]
            self.color_bottom_right = colors[0]
            self.color_top_left = colors[1]
            self.color_top_right = colors[1]
        elif len(colors) == 4:
            self.color_bottom_left = colors[0]
            self.color_bottom_right = colors[1]
            self.color_top_left = colors[2]
            self.color_top_right = colors[3]
        else:
            raise ValueError("Need 1, 2 or 4 colors.")

    @property
    def color_bottom_left(self):
        """The color in the bottom left corner."""
        return self.uniform_buffer.data["color_bottom_left"]

    @color_bottom_left.setter
    def color_bottom_left(self, color):
        self.uniform_buffer.data["color_bottom_left"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def color_bottom_right(self):
        """The color in the bottom right corner."""
        return self.uniform_buffer.data["color_bottom_right"]

    @color_bottom_right.setter
    def color_bottom_right(self, color):
        self.uniform_buffer.data["color_bottom_right"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def color_top_left(self):
        """The color in the top left corner."""
        return self.uniform_buffer.data["color_top_left"]

    @color_top_left.setter
    def color_top_left(self, color):
        self.uniform_buffer.data["color_top_left"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def color_top_right(self):
        """The color in the top right corner."""
        return self.uniform_buffer.data["color_top_right"]

    @color_top_right.setter
    def color_top_right(self, color):
        self.uniform_buffer.data["color_top_right"] = color
        self.uniform_buffer.update_range(0, 1)


class BackgroundImageMaterial(BackgroundMaterial):
    """A background material that displays an image. If map is a 2D
    texture view, it is used as a static background. If it is a cube
    texture view, (on a NxMx6 texture) it is used as a skybox.
    Use the Background object's transform to orient the image.
    """

    def __init__(self, map=None):
        super().__init__()
        self.map = map

    @property
    def map(self):
        """The texture map specifying the background image"""
        return self._map

    @map.setter
    def map(self, map):
        self._map = map



MACHINE_EPSILON = (
    7.0 / 3 - 4.0 / 3 - 1
)  # the difference between 1 and the smallest floating point number greater than 1


def clamp(x: float, left: float, right: float) -> float:
    return max(left, min(right, x))



# flake8: noqa

from ._base import Geometry
from ._box import BoxGeometry
from ._cylinder import CylinderGeometry
from ._plane import PlaneGeometry
from ._toroidal import KleinBottleGeometry, TorusKnotGeometry



import os
import io

from .. import Renderer, RenderFunctionRegistry

from ...objects import WorldObject
from ...cameras import Camera
from ...linalg import Matrix4, Vector3


registry = RenderFunctionRegistry()


def register_svg_render_function(wobject_cls, material_cls):
    """Decorator to register an SVG render function."""

    def _register_svg_render_function(f):
        registry.register(wobject_cls, material_cls, f)
        return f

    return _register_svg_render_function


class SvgRenderer(Renderer):
    """A renderer that generates an SVG."""

    def __init__(self, width, height, filename):
        self._width = width
        self._height = height

        # todo: also support writing to file-like object
        if filename.startswith("~"):
            filename = os.path.expanduser(filename)
        self._filename = filename

    def render(self, scene: WorldObject, camera: Camera):
        """Render the scene to a file."""

        # Ensure that matrices are up-to-date
        scene.update_matrix_world()
        camera.update_matrix_world()  # camera may not be a member of the scene
        camera.update_projection_matrix()

        # Get the sorted list of objects to render (guaranteed to be visible and having a material)
        proj_screen_matrix = Matrix4().multiply_matrices(
            camera.projection_matrix, camera.matrix_world_inverse
        )
        q = self.get_render_list(scene, proj_screen_matrix)

        # Init the svg file
        f = io.StringIO()
        f.write(
            f"<svg width='{self._width}' height='{self._height}' xmlns='http://www.w3.org/2000/svg'>\n"
        )

        # Render each object (that we can render)
        for wobject in q:
            renderfunc = registry.get_render_function(wobject)
            if renderfunc is not None:
                res = renderfunc(wobject)
                if isinstance(res, str):
                    f.write(res)
                elif isinstance(res, list):
                    for line in res:
                        f.write(line)

        # Finish the svg file and write actual file handle.
        f.write("\n</svg>\n")
        with open(self._filename, "wb") as f2:
            f2.write(f.getvalue().encode())

    def get_render_list(self, scene: WorldObject, proj_screen_matrix: Matrix4):
        """Given a scene object, get a list of objects to render."""

        # start by gathering everything that is visible and has a material
        q = []

        def visit(wobject):
            nonlocal q
            if wobject.visible and hasattr(wobject, "material"):
                q.append(wobject)

        scene.traverse(visit)

        # next, sort them from back-to-front
        def sort_func(wobject: WorldObject):
            z = (
                Vector3()
                .set_from_matrix_position(wobject.matrix_world)
                .apply_matrix4(proj_screen_matrix)
                .z
            )
            return wobject.render_order, z

        return list(sorted(q, key=sort_func))



from ._base import WorldObject


class Scene(WorldObject):
    """The scene is a WorldObject that represents the root of a scene graph.
    It hold certain attributes that relate to the scene, such as the background color,
    fog, and environment map. Camera's and lights can also be part of a scene.
    """


class Background(WorldObject):
    """An object representing a scene background.
    Can be e.g. a gradient, a static image or a skybox.
    """

    def __init__(self, material):
        super().__init__()
        self.material = material



from ..utils import array_from_shadertype
from ..resources import Buffer
from ._base import Material


class MeshBasicMaterial(Material):
    """A material for drawing geometries in a simple shaded (flat or
    wireframe) way. This material is not affected by lights.
    """

    uniform_type = {
        "color": ("float32", (4,)),
        "clim": ("float32", (2,)),
    }

    def __init__(self, **kwargs):
        super().__init__()

        self.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), usage="UNIFORM"
        )

        self._map = None
        self.color = 1, 1, 1, 1
        self.clim = 0, 1

        for argname, val in kwargs.items():
            if not hasattr(self, argname):
                raise AttributeError(f"No attribute '{argname}'")
            setattr(self, argname, val)

    def _wgpu_get_pick_info(self, pick_value):
        inst = pick_value[1]
        face = pick_value[2]
        coords = pick_value[3]
        coords = (coords & 0xFF0000) / 65536, (coords & 0xFF00) / 256, coords & 0xFF
        coords = coords[0] / 255, coords[1] / 255, coords[2] / 255
        return {"instance_index": inst, "face_index": face, "face_coords": coords}

    @property
    def color(self):
        """The uniform color of the mesh, as an rgba tuple.
        This value is ignored if a texture map is used.
        """
        return self.uniform_buffer.data["color"]

    @color.setter
    def color(self, color):
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_range(0, 1)

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate."""
        return self._map

    @map.setter
    def map(self, map):
        self._map = map

    @property
    def clim(self):
        """The contrast limits to apply to the map. Default (0, 1)"""
        return self.uniform_buffer.data["clim"]

    @clim.setter
    def clim(self, clim):
        self.uniform_buffer.data["clim"] = clim
        self.uniform_buffer.update_range(0, 1)


class MeshNormalMaterial(MeshBasicMaterial):
    """A material that maps the normal vectors to RGB colors."""


class MeshNormalLinesMaterial(MeshBasicMaterial):
    """A material that shows surface normals as lines sticking out of the mesh."""

    def _wgpu_get_pick_info(self, pick_value):
        return {}  # No picking for normal lines


# todo: MeshLambertMaterial(MeshBasicMaterial):
# A material for non-shiny surfaces, without specular highlights.


class MeshPhongMaterial(MeshBasicMaterial):
    """A material for shiny surfaces with specular highlights.

    The material uses a non-physically based Blinn-Phong model for
    calculating reflectance. Unlike the Lambertian model used in the
    MeshLambertMaterial this can simulate shiny surfaces with specular
    highlights (such as varnished wood).
    """


# todo: MeshStandardMaterial(MeshBasicMaterial):
# A standard physically based material, using Metallic-Roughness workflow.


# todo: MeshToonMaterial(MeshBasicMaterial):
# A cartoon-style mesh material.


class MeshSliceMaterial(MeshBasicMaterial):
    """A material that displays a slices of the mesh."""

    uniform_type = dict(
        color=("float32", 4),
        plane=("float32", 4),
        clim=("float32", 2),
        thickness=("float32",),
    )

    def __init__(self, plane=(0, 0, 1, 0), thickness=2.0, **kwargs):
        super().__init__(plane=plane, thickness=thickness, **kwargs)

    @property
    def plane(self):
        """The plane to slice at, represented with 4 floats ``(a, b, c, d)``,
        which make up the equation: ``ax + by + cz + d = 0`` The plane
        definition applies to the world space (of the scene).
        """
        return self.uniform_buffer.data["plane"]

    @plane.setter
    def plane(self, plane):
        self.uniform_buffer.data["plane"] = plane
        self.uniform_buffer.update_range(0, 1)

    @property
    def thickness(self):
        """The thickness of the line to draw the edge of the mesh."""
        return self.uniform_buffer.data["thickness"]

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = thickness
        self.uniform_buffer.update_range(0, 1)



# flake8: noqa

from ._base import WorldObject
from ._group import Group
from ._scene import Scene, Background
from ._points import Points
from ._line import Line
from ._mesh import Mesh, InstancedMesh
from ._volume import Volume



import numpy as np

from ..resources import Buffer
from ._base import Geometry


class PlaneGeometry(Geometry):
    def __init__(self, width=1, height=1, width_segments=1, height_segments=1):
        super().__init__()
        nx, ny = width_segments + 1, height_segments + 1

        x = np.linspace(-width / 2, width / 2, nx, dtype=np.float32)
        y = np.linspace(-height / 2, height / 2, ny, dtype=np.float32)
        xx, yy = np.meshgrid(x, y)
        xx, yy = xx.flatten(), yy.flatten()
        positions = np.column_stack([xx, yy, np.zeros_like(xx)])

        x = np.linspace(0, 1, nx, dtype=np.float32)
        y = np.linspace(0, 1, ny, dtype=np.float32)
        xx, yy = np.meshgrid(x, y)
        texcoords = np.column_stack([xx.flat, yy.flat])

        indices = np.zeros((height_segments, width_segments, 6), np.uint32)

        for y in range(height_segments):
            for x in range(width_segments):
                i = x + y * nx
                indices[y, x, :] = i, i + 1, i + nx, i + nx, i + 1, i + nx + 1

        indices = indices.ravel()

        self.positions = Buffer(positions, usage="vertex|storage")
        self.texcoords = Buffer(texcoords, usage="vertex|storage")
        self.index = Buffer(indices, usage="index|storage")



from ._base import Material
from ..utils import array_from_shadertype
from ..resources import Buffer


class LineMaterial(Material):
    """The default material to draw lines."""

    uniform_type = dict(
        color=("float32", 4),
        thickness=("float32",),
    )

    def __init__(self, color=(1, 1, 1, 1), thickness=2.0, vertex_colors=False):
        super().__init__()

        self.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), usage="UNIFORM"
        )
        self.set_color(color)
        self.set_thickness(thickness)

        self._vertex_colors = vertex_colors

    def _wgpu_get_pick_info(self, pick_value):
        # The instance is zero while renderer doesn't support instancing
        instance = pick_value[1]
        vertex = pick_value[2]
        vertex_sub = pick_value[3] / 1048576
        return {"instance_index": instance, "vertex_index": vertex + vertex_sub}

    @property
    def color(self):
        return self.uniform_buffer.data["color"]

    def set_color(self, color):
        self.uniform_buffer.data["color"] = tuple(color)
        self.uniform_buffer.update_range(0, 1)

    @property
    def vertex_colors(self):
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, value):
        if value != self._vertex_colors:
            self._vertex_colors = value
            self._bump_rev()

    # todo: thickness? maybe rename to width?
    @property
    def thickness(self):
        """The line thickness expressed in logical pixels."""
        return self.uniform_buffer.data["thickness"]

    def set_thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = thickness
        self.uniform_buffer.update_range(0, 1)


class LineThinMaterial(LineMaterial):
    """A simple line, drawn with line_strip primitives that has a width
    of one physical pixel. The thickness is ignored.
    """


class LineThinSegmentMaterial(LineMaterial):
    """Simple line segments, drawn with line primitives that has a width
    of one physical pixel. The thickness is ignored.
    """


class LineSegmentMaterial(LineMaterial):
    """A material that renders line segments between each two subsequent points."""


class LineArrowMaterial(LineSegmentMaterial):
    """A material that renders line segments that look like little vectors."""



import re

from setuptools import find_packages, setup


NAME = "pygfx"
SUMMARY = "A threejs-like render engine based on wgpu"

with open(f"{NAME}/__init__.py") as fh:
    VERSION = re.search(r"__version__ = \"(.*?)\"", fh.read()).group(1)


runtime_deps = [
    "numpy",
    "wgpu>=0.5.4",
    "Jinja2",
]


setup(
    name=NAME,
    version=VERSION,
    packages=find_packages(
        exclude=["tests", "tests.*", "examples", "examples.*", "exp", "exp.*"]
    ),
    python_requires=">=3.6.0",
    install_requires=runtime_deps,
    license="BSD 2-Clause",
    description=SUMMARY,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Almar Klein",
    author_email="almar.klein@gmail.com",
    url="https://github.com/pygfx/pygfx",
    data_files=[("", ["LICENSE"])],
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



class Vector4:
    def __init__(self, x: float = 0, y: float = 0, z: float = 0, w: float = 0) -> None:
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def __repr__(self) -> str:
        return f"Vector4({self.x}, {self.y}, {self.z}, {self.w})"

    def set(self, x: float, y: float, z: float, w: float) -> "Vector4":
        self.x = x
        self.y = y
        self.z = z
        self.w = w
        return self

    def apply_matrix4(self, m: "Matrix4") -> "Vector4":
        x = self.x
        y = self.y
        z = self.z
        w = self.w
        e = m.elements

        self.x = e[0] * x + e[4] * y + e[8] * z + e[12] * w
        self.y = e[1] * x + e[5] * y + e[9] * z + e[13] * w
        self.z = e[2] * x + e[6] * y + e[10] * z + e[14] * w
        self.w = e[3] * x + e[7] * y + e[11] * z + e[15] * w

        return self



from typing import Tuple

from ..linalg import Vector3, Matrix4, Quaternion
from ._orbit import get_screen_vectors_in_world_cords


class PanZoomControls:
    def __init__(
        self,
        eye: Vector3 = None,
        target: Vector3 = None,
        up: Vector3 = None,
        zoom: float = 1.0,
        min_zoom: float = 0.0001,
    ) -> None:
        self.rotation = Quaternion()
        if eye is None:
            eye = Vector3(50.0, 50.0, 50.0)
        if target is None:
            target = Vector3()
        if up is None:
            up = Vector3(0.0, 1.0, 0.0)
        self.zoom_value = zoom
        self.min_zoom = min_zoom

        # State info used during a pan or rotate operation
        self._pan_info = None

        # Temp objects (to avoid garbage collection)
        self._m = Matrix4()
        self._v = Vector3()

        # Initialize orientation
        self.look_at(eye, target, up)

    def look_at(self, eye: Vector3, target: Vector3, up: Vector3) -> "PanZoomControls":
        self.distance = eye.distance_to(target)
        self.target = target
        self.up = up
        self.rotation.set_from_rotation_matrix(self._m.look_at(eye, target, up))
        return self

    def pan(self, vec3: Vector3) -> "PanZoomControls":
        """Pan in 3D world coordinates."""
        self.target.add(vec3)
        return self

    def pan_start(
        self,
        pos: Tuple[float, float],
        canvas_size: Tuple[float, float],
        camera: "Camera",
    ) -> "PanZoomControls":
        # Using this function may be a bit overkill. We can also simply
        # get the ortho cameras world_size (camera.visible_world_size).
        # However, now the panzoom controls work with a perspecive camera ...
        vecx, vecy = get_screen_vectors_in_world_cords(self.target, canvas_size, camera)
        self._pan_info = {"last": pos, "vecx": vecx, "vecy": vecy}
        return self

    def pan_stop(self) -> "PanZoomControls":
        self._pan_info = None
        return self

    def pan_move(self, pos: Tuple[float, float]) -> "PanZoomControls":
        """Pan the camera, based on a (2D) screen location. Call pan_start first."""
        if self._pan_info is None:
            return
        delta = tuple((pos[i] - self._pan_info["last"][i]) for i in range(2))
        self.pan(
            self._pan_info["vecx"]
            .clone()
            .multiply_scalar(-delta[0])
            .add_scaled_vector(self._pan_info["vecy"], +delta[1])
        )
        self._pan_info["last"] = pos
        return self

    def zoom(self, multiplier: float) -> "PanZoomControls":
        self.zoom_value = max(self.min_zoom, float(multiplier) * self.zoom_value)
        return self

    def zoom_to_point(
        self,
        multiplier: float,
        pos: Tuple[float, float],
        canvas_size: Tuple[float, float],
        camera: "Camera",
    ) -> "PanZoomControls":

        # Apply zoom
        zoom_old = self.zoom_value
        self.zoom(multiplier)
        zoom_ratio = zoom_old / self.zoom_value  # usually == multiplier

        # Now pan such that what was previously under the mouse is again under the mouse.
        vecx, vecy = get_screen_vectors_in_world_cords(self.target, canvas_size, camera)
        delta = tuple(pos[i] - canvas_size[i] / 2 for i in (0, 1))
        delta1 = vecx.multiply_scalar(delta[0]).add(vecy.multiply_scalar(-delta[1]))
        delta2 = delta1.clone().multiply_scalar(zoom_ratio)
        self.pan(delta1.sub(delta2))
        return self

    def get_view(self) -> Tuple[Vector3, Vector3, float]:
        self._v.set(0, 0, self.distance).apply_quaternion(self.rotation).add(
            self.target
        )
        return self.rotation, self._v, self.zoom_value

    def update_camera(self, camera: "Camera") -> "PanZoomControls":
        rot, pos, zoom = self.get_view()
        camera.rotation.copy(rot)
        camera.position.copy(pos)
        camera.zoom = zoom
        return self



from math import cos, sin

from .vector3 import Vector3


class Matrix4:
    def __init__(
        self,
        n11: float = 1,
        n12: float = 0,
        n13: float = 0,
        n14: float = 0,
        n21: float = 0,
        n22: float = 1,
        n23: float = 0,
        n24: float = 0,
        n31: float = 0,
        n32: float = 0,
        n33: float = 1,
        n34: float = 0,
        n41: float = 0,
        n42: float = 0,
        n43: float = 0,
        n44: float = 1,
    ) -> None:
        self.elements = [
            n11,
            n12,
            n13,
            n14,
            n21,
            n22,
            n23,
            n24,
            n31,
            n32,
            n33,
            n34,
            n41,
            n42,
            n43,
            n44,
        ]

    def __repr__(self) -> str:
        return f"Matrix4({self.elements[0]}, {self.elements[1]}, {self.elements[2]}, {self.elements[3]}, {self.elements[4]}, {self.elements[5]}, {self.elements[6]}, {self.elements[7]}, {self.elements[8]}, {self.elements[9]}, {self.elements[10]}, {self.elements[11]}, {self.elements[12]}, {self.elements[13]}, {self.elements[14]}, {self.elements[15]})"

    def set(
        self,
        n11: float,
        n12: float,
        n13: float,
        n14: float,
        n21: float,
        n22: float,
        n23: float,
        n24: float,
        n31: float,
        n32: float,
        n33: float,
        n34: float,
        n41: float,
        n42: float,
        n43: float,
        n44: float,
    ) -> "Matrix4":
        te = self.elements

        te[0] = n11
        te[4] = n12
        te[8] = n13
        te[12] = n14
        te[1] = n21
        te[5] = n22
        te[9] = n23
        te[13] = n24
        te[2] = n31
        te[6] = n32
        te[10] = n33
        te[14] = n34
        te[3] = n41
        te[7] = n42
        te[11] = n43
        te[15] = n44

        return self

    def identity(self) -> "Matrix4":
        self.set(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1)
        return self

    def clone(self) -> "Matrix4":
        return Matrix4().from_array(self.elements)

    def copy(self, m: "Matrix4") -> "Matrix4":
        te = self.elements
        me = m.elements

        te[0] = me[0]
        te[1] = me[1]
        te[2] = me[2]
        te[3] = me[3]
        te[4] = me[4]
        te[5] = me[5]
        te[6] = me[6]
        te[7] = me[7]
        te[8] = me[8]
        te[9] = me[9]
        te[10] = me[10]
        te[11] = me[11]
        te[12] = me[12]
        te[13] = me[13]
        te[14] = me[14]
        te[15] = me[15]

        return self

    def copy_position(self, m: "Matrix4") -> "Matrix4":
        te = self.elements
        me = m.elements

        te[12] = me[12]
        te[13] = me[13]
        te[14] = me[14]

        return self

    def extract_basis(
        self, x_axis: "Vector3", y_axis: "Vector3", z_axis: "Vector3"
    ) -> "Matrix4":
        x_axis.set_from_matrix_column(self, 0)
        y_axis.set_from_matrix_column(self, 1)
        z_axis.set_from_matrix_column(self, 2)
        return self

    def make_basis(
        self, x_axis: "Vector3", y_axis: "Vector3", z_axis: "Vector3"
    ) -> "Matrix4":
        self.set(
            x_axis.x,
            y_axis.x,
            z_axis.x,
            0,
            x_axis.y,
            y_axis.y,
            z_axis.y,
            0,
            x_axis.z,
            y_axis.z,
            z_axis.z,
            0,
            0,
            0,
            0,
            1,
        )

        return self

    def extract_rotation(self, m: "Matrix4") -> "Matrix4":
        # this method does not support reflection matrices
        te = self.elements
        me = m.elements

        scale_x = 1 / _tmp_vector.set_from_matrix_column(m, 0).length()
        scale_y = 1 / _tmp_vector.set_from_matrix_column(m, 1).length()
        scale_z = 1 / _tmp_vector.set_from_matrix_column(m, 2).length()

        te[0] = me[0] * scale_x
        te[1] = me[1] * scale_x
        te[2] = me[2] * scale_x
        te[3] = 0
        te[4] = me[4] * scale_y
        te[5] = me[5] * scale_y
        te[6] = me[6] * scale_y
        te[7] = 0
        te[8] = me[8] * scale_z
        te[9] = me[9] * scale_z
        te[10] = me[10] * scale_z
        te[11] = 0
        te[12] = 0
        te[13] = 0
        te[14] = 0
        te[15] = 1

        return self

    def make_rotation_from_euler(self, euler: "Euler") -> "Matrix4":
        from .euler import Euler

        te = self.elements
        x = euler.x
        y = euler.y
        z = euler.z
        a = cos(x)
        b = sin(x)
        c = cos(y)
        d = sin(y)
        e = cos(z)
        f = sin(z)

        if euler.order == Euler.RotationOrders.XYZ:
            ae = a * e
            af = a * f
            be = b * e
            bf = b * f

            te[0] = c * e
            te[4] = -c * f
            te[8] = d

            te[1] = af + be * d
            te[5] = ae - bf * d
            te[9] = -b * c

            te[2] = bf - ae * d
            te[6] = be + af * d
            te[10] = a * c
        elif euler.order == Euler.RotationOrders.YXZ:
            ce = c * e
            cf = c * f
            de = d * e
            df = d * f

            te[0] = ce + df * b
            te[4] = de * b - cf
            te[8] = a * d

            te[1] = a * f
            te[5] = a * e
            te[9] = -b

            te[2] = cf * b - de
            te[6] = df + ce * b
            te[10] = a * c
        elif euler.order == Euler.RotationOrders.ZXY:
            ce = c * e
            cf = c * f
            de = d * e
            df = d * f

            te[0] = ce - df * b
            te[4] = -a * f
            te[8] = de + cf * b

            te[1] = cf + de * b
            te[5] = a * e
            te[9] = df - ce * b

            te[2] = -a * d
            te[6] = b
            te[10] = a * c
        elif euler.order == Euler.RotationOrders.ZYX:
            ae = a * e
            af = a * f
            be = b * e
            bf = b * f

            te[0] = c * e
            te[4] = be * d - af
            te[8] = ae * d + bf

            te[1] = c * f
            te[5] = bf * d + ae
            te[9] = af * d - be

            te[2] = -d
            te[6] = b * c
            te[10] = a * c
        elif euler.order == Euler.RotationOrders.YZX:
            ac = a * c
            ad = a * d
            bc = b * c
            bd = b * d

            te[0] = c * e
            te[4] = bd - ac * f
            te[8] = bc * f + ad

            te[1] = f
            te[5] = a * e
            te[9] = -b * e

            te[2] = -d * e
            te[6] = ad * f + bc
            te[10] = ac - bd * f
        elif euler.order == Euler.RotationOrders.XZY:
            ac = a * c
            ad = a * d
            bc = b * c
            bd = b * d

            te[0] = c * e
            te[4] = -f
            te[8] = d * e

            te[1] = ac * f + bd
            te[5] = a * e
            te[9] = ad * f - bc

            te[2] = bc * f - ad
            te[6] = b * e
            te[10] = bd * f + ac

        # bottom row
        te[3] = 0
        te[7] = 0
        te[11] = 0

        # last column
        te[12] = 0
        te[13] = 0
        te[14] = 0
        te[15] = 1

        return self

    def make_rotation_from_quaternion(self, q: "Quaternion") -> "Matrix4":
        return self.compose(_zero, q, _one)

    def look_at(self, eye: "Vector3", target: "Vector3", up: "Vector3") -> "Matrix4":
        te = self.elements

        _tmp_vector.sub_vectors(eye, target)

        if _tmp_vector.length_sq() == 0:
            # eye and target are in the same position
            _tmp_vector.z = 1

        _tmp_vector.normalize()
        _tmp_vector2.cross_vectors(up, _tmp_vector)

        if _tmp_vector2.length_sq() == 0:
            # up and z are parallel
            if abs(up.z) == 1:
                _tmp_vector.x += 0.0001
            else:
                _tmp_vector.z += 0.0001

            _tmp_vector.normalize()
            _tmp_vector2.cross_vectors(up, _tmp_vector)

        _tmp_vector2.normalize()
        _tmp_vector3.cross_vectors(_tmp_vector, _tmp_vector2)

        te[0] = _tmp_vector2.x
        te[4] = _tmp_vector3.x
        te[8] = _tmp_vector.x
        te[1] = _tmp_vector2.y
        te[5] = _tmp_vector3.y
        te[9] = _tmp_vector.y
        te[2] = _tmp_vector2.z
        te[6] = _tmp_vector3.z
        te[10] = _tmp_vector.z

        return self

    def multiply(self, m: "Matrix4") -> "Matrix4":
        return self.multiply_matrices(self, m)

    def premultiply(self, m: "Matrix4") -> "Matrix4":
        return self.multiply_matrices(m, self)

    def multiply_matrices(self, a: "Matrix4", b: "Matrix4") -> "Matrix4":
        ae = a.elements
        be = b.elements
        te = self.elements

        a11 = ae[0]
        a12 = ae[4]
        a13 = ae[8]
        a14 = ae[12]
        a21 = ae[1]
        a22 = ae[5]
        a23 = ae[9]
        a24 = ae[13]
        a31 = ae[2]
        a32 = ae[6]
        a33 = ae[10]
        a34 = ae[14]
        a41 = ae[3]
        a42 = ae[7]
        a43 = ae[11]
        a44 = ae[15]

        b11 = be[0]
        b12 = be[4]
        b13 = be[8]
        b14 = be[12]
        b21 = be[1]
        b22 = be[5]
        b23 = be[9]
        b24 = be[13]
        b31 = be[2]
        b32 = be[6]
        b33 = be[10]
        b34 = be[14]
        b41 = be[3]
        b42 = be[7]
        b43 = be[11]
        b44 = be[15]

        te[0] = a11 * b11 + a12 * b21 + a13 * b31 + a14 * b41
        te[4] = a11 * b12 + a12 * b22 + a13 * b32 + a14 * b42
        te[8] = a11 * b13 + a12 * b23 + a13 * b33 + a14 * b43
        te[12] = a11 * b14 + a12 * b24 + a13 * b34 + a14 * b44

        te[1] = a21 * b11 + a22 * b21 + a23 * b31 + a24 * b41
        te[5] = a21 * b12 + a22 * b22 + a23 * b32 + a24 * b42
        te[9] = a21 * b13 + a22 * b23 + a23 * b33 + a24 * b43
        te[13] = a21 * b14 + a22 * b24 + a23 * b34 + a24 * b44

        te[2] = a31 * b11 + a32 * b21 + a33 * b31 + a34 * b41
        te[6] = a31 * b12 + a32 * b22 + a33 * b32 + a34 * b42
        te[10] = a31 * b13 + a32 * b23 + a33 * b33 + a34 * b43
        te[14] = a31 * b14 + a32 * b24 + a33 * b34 + a34 * b44

        te[3] = a41 * b11 + a42 * b21 + a43 * b31 + a44 * b41
        te[7] = a41 * b12 + a42 * b22 + a43 * b32 + a44 * b42
        te[11] = a41 * b13 + a42 * b23 + a43 * b33 + a44 * b43
        te[15] = a41 * b14 + a42 * b24 + a43 * b34 + a44 * b44

        return self

    def multiply_scalar(self, s: float) -> "Matrix4":
        te = self.elements

        te[0] *= s
        te[4] *= s
        te[8] *= s
        te[12] *= s
        te[1] *= s
        te[5] *= s
        te[9] *= s
        te[13] *= s
        te[2] *= s
        te[6] *= s
        te[10] *= s
        te[14] *= s
        te[3] *= s
        te[7] *= s
        te[11] *= s
        te[15] *= s

        return self

    def determinant(self) -> float:
        te = self.elements

        n11 = te[0]
        n12 = te[4]
        n13 = te[8]
        n14 = te[12]
        n21 = te[1]
        n22 = te[5]
        n23 = te[9]
        n24 = te[13]
        n31 = te[2]
        n32 = te[6]
        n33 = te[10]
        n34 = te[14]
        n41 = te[3]
        n42 = te[7]
        n43 = te[11]
        n44 = te[15]

        # TODO: make this more efficient
        # (based on http:#www.euclideanspace.com/maths/algebra/matrix/functions/inverse/fourD/index.htm)

        return (
            n41
            * (
                +n14 * n23 * n32
                - n13 * n24 * n32
                - n14 * n22 * n33
                + n12 * n24 * n33
                + n13 * n22 * n34
                - n12 * n23 * n34
            )
            + n42
            * (
                +n11 * n23 * n34
                - n11 * n24 * n33
                + n14 * n21 * n33
                - n13 * n21 * n34
                + n13 * n24 * n31
                - n14 * n23 * n31
            )
            + n43
            * (
                +n11 * n24 * n32
                - n11 * n22 * n34
                - n14 * n21 * n32
                + n12 * n21 * n34
                + n14 * n22 * n31
                - n12 * n24 * n31
            )
            + n44
            * (
                -n13 * n22 * n31
                - n11 * n23 * n32
                + n11 * n22 * n33
                + n13 * n21 * n32
                - n12 * n21 * n33
                + n12 * n23 * n31
            )
        )

    def transpose(self) -> "Matrix4":
        te = self.elements
        tmp = None

        tmp = te[1]
        te[1] = te[4]
        te[4] = tmp

        tmp = te[2]
        te[2] = te[8]
        te[8] = tmp

        tmp = te[6]
        te[6] = te[9]
        te[9] = tmp

        tmp = te[3]
        te[3] = te[12]
        te[12] = tmp

        tmp = te[7]
        te[7] = te[13]
        te[13] = tmp

        tmp = te[11]
        te[11] = te[14]
        te[14] = tmp

        return self

    def set_position(self, v: "Vector3") -> "Matrix4":
        te = self.elements
        te[12] = v.x
        te[13] = v.y
        te[14] = v.z
        return self

    def set_position_xyz(self, x: float, y: float, z: float) -> "Matrix4":
        te = self.elements
        te[12] = x
        te[13] = y
        te[14] = z
        return self

    def get_inverse(self, m: "Matrix4", throw_on_degenerate: bool = True) -> "Matrix4":
        te = self.elements
        me = m.elements

        n11 = me[0]
        n21 = me[1]
        n31 = me[2]
        n41 = me[3]
        n12 = me[4]
        n22 = me[5]
        n32 = me[6]
        n42 = me[7]
        n13 = me[8]
        n23 = me[9]
        n33 = me[10]
        n43 = me[11]
        n14 = me[12]
        n24 = me[13]
        n34 = me[14]
        n44 = me[15]

        t11 = (
            n23 * n34 * n42
            - n24 * n33 * n42
            + n24 * n32 * n43
            - n22 * n34 * n43
            - n23 * n32 * n44
            + n22 * n33 * n44
        )
        t12 = (
            n14 * n33 * n42
            - n13 * n34 * n42
            - n14 * n32 * n43
            + n12 * n34 * n43
            + n13 * n32 * n44
            - n12 * n33 * n44
        )
        t13 = (
            n13 * n24 * n42
            - n14 * n23 * n42
            + n14 * n22 * n43
            - n12 * n24 * n43
            - n13 * n22 * n44
            + n12 * n23 * n44
        )
        t14 = (
            n14 * n23 * n32
            - n13 * n24 * n32
            - n14 * n22 * n33
            + n12 * n24 * n33
            + n13 * n22 * n34
            - n12 * n23 * n34
        )

        det = n11 * t11 + n21 * t12 + n31 * t13 + n41 * t14
        if det == 0:
            if throw_on_degenerate:
                raise ValueError("matrix determinant is zero, cannot invert")
            else:
                return self.identity()

        det_inv = 1 / det

        te[0] = t11 * det_inv
        te[1] = (
            n24 * n33 * n41
            - n23 * n34 * n41
            - n24 * n31 * n43
            + n21 * n34 * n43
            + n23 * n31 * n44
            - n21 * n33 * n44
        ) * det_inv
        te[2] = (
            n22 * n34 * n41
            - n24 * n32 * n41
            + n24 * n31 * n42
            - n21 * n34 * n42
            - n22 * n31 * n44
            + n21 * n32 * n44
        ) * det_inv
        te[3] = (
            n23 * n32 * n41
            - n22 * n33 * n41
            - n23 * n31 * n42
            + n21 * n33 * n42
            + n22 * n31 * n43
            - n21 * n32 * n43
        ) * det_inv

        te[4] = t12 * det_inv
        te[5] = (
            n13 * n34 * n41
            - n14 * n33 * n41
            + n14 * n31 * n43
            - n11 * n34 * n43
            - n13 * n31 * n44
            + n11 * n33 * n44
        ) * det_inv
        te[6] = (
            n14 * n32 * n41
            - n12 * n34 * n41
            - n14 * n31 * n42
            + n11 * n34 * n42
            + n12 * n31 * n44
            - n11 * n32 * n44
        ) * det_inv
        te[7] = (
            n12 * n33 * n41
            - n13 * n32 * n41
            + n13 * n31 * n42
            - n11 * n33 * n42
            - n12 * n31 * n43
            + n11 * n32 * n43
        ) * det_inv

        te[8] = t13 * det_inv
        te[9] = (
            n14 * n23 * n41
            - n13 * n24 * n41
            - n14 * n21 * n43
            + n11 * n24 * n43
            + n13 * n21 * n44
            - n11 * n23 * n44
        ) * det_inv
        te[10] = (
            n12 * n24 * n41
            - n14 * n22 * n41
            + n14 * n21 * n42
            - n11 * n24 * n42
            - n12 * n21 * n44
            + n11 * n22 * n44
        ) * det_inv
        te[11] = (
            n13 * n22 * n41
            - n12 * n23 * n41
            - n13 * n21 * n42
            + n11 * n23 * n42
            + n12 * n21 * n43
            - n11 * n22 * n43
        ) * det_inv

        te[12] = t14 * det_inv
        te[13] = (
            n13 * n24 * n31
            - n14 * n23 * n31
            + n14 * n21 * n33
            - n11 * n24 * n33
            - n13 * n21 * n34
            + n11 * n23 * n34
        ) * det_inv
        te[14] = (
            n14 * n22 * n31
            - n12 * n24 * n31
            - n14 * n21 * n32
            + n11 * n24 * n32
            + n12 * n21 * n34
            - n11 * n22 * n34
        ) * det_inv
        te[15] = (
            n12 * n23 * n31
            - n13 * n22 * n31
            + n13 * n21 * n32
            - n11 * n23 * n32
            - n12 * n21 * n33
            + n11 * n22 * n33
        ) * det_inv

        return self

    def scale(self, v: "Vector3") -> "Matrix4":
        te = self.elements
        x = v.x
        y = v.y
        z = v.z

        te[0] *= x
        te[4] *= y
        te[8] *= z
        te[1] *= x
        te[5] *= y
        te[9] *= z
        te[2] *= x
        te[6] *= y
        te[10] *= z
        te[3] *= x
        te[7] *= y
        te[11] *= z

        return self

    def get_max_scale_on_axis(self) -> float:
        te = self.elements

        scale_x_sq = te[0] * te[0] + te[1] * te[1] + te[2] * te[2]
        scale_y_sq = te[4] * te[4] + te[5] * te[5] + te[6] * te[6]
        scale_z_sq = te[8] * te[8] + te[9] * te[9] + te[10] * te[10]

        return max(scale_x_sq, scale_y_sq, scale_z_sq) ** 0.5

    def make_translation(self, x: float, y: float, z: float) -> "Matrix4":
        self.set(1, 0, 0, x, 0, 1, 0, y, 0, 0, 1, z, 0, 0, 0, 1)
        return self

    def make_rotation_x(self, theta: float) -> "Matrix4":
        c = cos(theta)
        s = sin(theta)
        self.set(1, 0, 0, 0, 0, c, -s, 0, 0, s, c, 0, 0, 0, 0, 1)
        return self

    def make_rotation_y(self, theta: float) -> "Matrix4":
        c = cos(theta)
        s = sin(theta)
        self.set(c, 0, s, 0, 0, 1, 0, 0, -s, 0, c, 0, 0, 0, 0, 1)
        return self

    def make_rotation_z(self, theta: float) -> "Matrix4":
        c = cos(theta)
        s = sin(theta)
        self.set(c, -s, 0, 0, s, c, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1)
        return self

    def make_rotation_axis(self, axis: "Vector3", angle: float) -> "Matrix4":
        # Based on http:#www.gamedev.net/reference/articles/article1199.asp
        c = cos(angle)
        s = sin(angle)
        t = 1 - c
        x = axis.x
        y = axis.y
        z = axis.z
        tx = t * x
        ty = t * y
        self.set(
            tx * x + c,
            tx * y - s * z,
            tx * z + s * y,
            0,
            tx * y + s * z,
            ty * y + c,
            ty * z - s * x,
            0,
            tx * z - s * y,
            ty * z + s * x,
            t * z * z + c,
            0,
            0,
            0,
            0,
            1,
        )
        return self

    def make_scale(self, x: float, y: float, z: float) -> "Matrix4":
        self.set(x, 0, 0, 0, 0, y, 0, 0, 0, 0, z, 0, 0, 0, 0, 1)
        return self

    def make_shear(self, x: float, y: float, z: float) -> "Matrix4":
        self.set(1, y, z, 0, x, 1, z, 0, x, y, 1, 0, 0, 0, 0, 1)
        return self

    def compose(
        self, position: "Vector3", quaternion: "Quaternion", scale: "Vector3"
    ) -> "Matrix4":
        te = self.elements

        x = quaternion.x
        y = quaternion.y
        z = quaternion.z
        w = quaternion.w
        x2 = x + x
        y2 = y + y
        z2 = z + z
        xx = x * x2
        xy = x * y2
        xz = x * z2
        yy = y * y2
        yz = y * z2
        zz = z * z2
        wx = w * x2
        wy = w * y2
        wz = w * z2

        sx = scale.x
        sy = scale.y
        sz = scale.z

        te[0] = (1 - (yy + zz)) * sx
        te[1] = (xy + wz) * sx
        te[2] = (xz - wy) * sx
        te[3] = 0

        te[4] = (xy - wz) * sy
        te[5] = (1 - (xx + zz)) * sy
        te[6] = (yz + wx) * sy
        te[7] = 0

        te[8] = (xz + wy) * sz
        te[9] = (yz - wx) * sz
        te[10] = (1 - (xx + yy)) * sz
        te[11] = 0

        te[12] = position.x
        te[13] = position.y
        te[14] = position.z
        te[15] = 1

        return self

    def decompose(
        self, position: "Vector3", quaternion: "Quaternion", scale: "Vector3"
    ) -> "Matrix4":
        te = self.elements

        sx = _tmp_vector.set(te[0], te[1], te[2]).length()
        sy = _tmp_vector.set(te[4], te[5], te[6]).length()
        sz = _tmp_vector.set(te[8], te[9], te[10]).length()

        # if determine is negative, we need to invert one scale
        det = self.determinant()
        if det < 0:
            sx = -sx

        position.x = te[12]
        position.y = te[13]
        position.z = te[14]

        # scale the rotation part
        _tmp_matrix4.copy(self)

        inv_sx = 1 / sx
        inv_sy = 1 / sy
        inv_sz = 1 / sz

        _tmp_matrix4.elements[0] *= inv_sx
        _tmp_matrix4.elements[1] *= inv_sx
        _tmp_matrix4.elements[2] *= inv_sx

        _tmp_matrix4.elements[4] *= inv_sy
        _tmp_matrix4.elements[5] *= inv_sy
        _tmp_matrix4.elements[6] *= inv_sy

        _tmp_matrix4.elements[8] *= inv_sz
        _tmp_matrix4.elements[9] *= inv_sz
        _tmp_matrix4.elements[10] *= inv_sz

        quaternion.set_from_rotation_matrix(_tmp_matrix4)

        scale.x = sx
        scale.y = sy
        scale.z = sz

        return self

    def make_perspective(
        self,
        left: float,
        right: float,
        top: float,
        bottom: float,
        near: float,
        far: float,
    ) -> "Matrix4":
        te = self.elements
        x = 2 * near / (right - left)
        y = 2 * near / (top - bottom)

        a = (right + left) / (right - left)
        b = (top + bottom) / (top - bottom)
        c = -(far + near) / (far - near)
        d = -2 * far * near / (far - near)

        te[0] = x
        te[4] = 0
        te[8] = a
        te[12] = 0
        te[1] = 0
        te[5] = y
        te[9] = b
        te[13] = 0
        te[2] = 0
        te[6] = 0
        te[10] = c
        te[14] = d
        te[3] = 0
        te[7] = 0
        te[11] = -1
        te[15] = 0

        return self

    def make_orthographic(
        self,
        left: float,
        right: float,
        top: float,
        bottom: float,
        near: float,
        far: float,
    ) -> "Matrix4":
        te = self.elements
        w = 1.0 / (right - left)
        h = 1.0 / (top - bottom)
        p = 1.0 / (far - near)

        x = (right + left) * w
        y = (top + bottom) * h
        z = (far + near) * p

        te[0] = 2 * w
        te[4] = 0
        te[8] = 0
        te[12] = -x
        te[1] = 0
        te[5] = 2 * h
        te[9] = 0
        te[13] = -y
        te[2] = 0
        te[6] = 0
        te[10] = -2 * p
        te[14] = -z
        te[3] = 0
        te[7] = 0
        te[11] = 0
        te[15] = 1

        return self

    def equals(self, matrix: "Matrix4") -> bool:
        te = self.elements
        me = matrix.elements
        for i in range(16):
            if te[i] != me[i]:
                return False
        return True

    def __eq__(self, other: "Matrix4") -> bool:
        return isinstance(other, Matrix4) and self.equals(other)

    def from_array(self, array: list, offset: int = 0) -> "Matrix4":
        for i in range(16):
            self.elements[i] = array[i + offset]
        return self

    def to_array(self, array: list = None, offset: int = 0) -> list:
        if array is None:
            array = []
        padding = offset + 16 - len(array)
        if padding > 0:
            array.extend((None for _ in range(padding)))
        for i in range(16):
            array[i + offset] = self.elements[i]
        return array


_zero = Vector3(0, 0, 0)
_one = Vector3(1, 1, 1)
_tmp_vector = Vector3()
_tmp_vector2 = Vector3()
_tmp_vector3 = Vector3()
_tmp_matrix4 = Matrix4()



import jinja2
import numpy as np

jinja_env = jinja2.Environment(
    block_start_string="{$",
    block_end_string="$}",
    variable_start_string="{{",
    variable_end_string="}}",
    line_statement_prefix="$$",
    undefined=jinja2.StrictUndefined,
)


class BaseShader:
    """Base shader object to compose and template shaders using jinja2."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self._uniform_codes = {}

    def __setitem__(self, key, value):
        self.kwargs[key] = value

    def __getitem__(self, key):
        return self.kwargs[key]

    def get_definitions(self):
        return "\n".join(self._uniform_codes.values())

    def get_code(self):
        raise NotImplementedError()

    def generate_wgsl(self, **kwargs):
        code = self.get_code()
        t = jinja_env.from_string(code)

        variables = self.kwargs.copy()
        variables.update(kwargs)

        try:
            return t.render(**variables)
        except jinja2.UndefinedError as err:
            msg = f"Canot compose shader: {err.message}"
        raise ValueError(msg)  # don't raise within handler to avoid recursive tb

    def define_uniform(self, bindgroup, index, name, struct):

        structname = "Struct_" + name
        code = f"""
        [[block]]
        struct {structname} {{
        """.rstrip()

        if isinstance(struct, dict):
            dtype_struct = np.dtype([(key,) + val for key, val in struct.items()])
        elif isinstance(struct, np.dtype):
            if struct.fields is None:
                raise TypeError(f"define_uniform() needs a structured dtype")
            dtype_struct = struct
        else:
            raise TypeError(f"Unsupported struct type {struct.__class__.__name__}")

        for fieldname, (dtype, offset) in dtype_struct.fields.items():
            # Resolve primitive type
            primitive_type = dtype.base.name
            primitive_type = primitive_type.replace("float", "f")
            primitive_type = primitive_type.replace("uint", "u")
            primitive_type = primitive_type.replace("int", "i")
            # Resolve actual type (only scalar, vec, mat)
            shape = dtype.shape
            if shape == () or shape == (1,):
                wgsl_type = primitive_type
            elif len(shape) == 1:
                wgsl_type = f"vec{shape[0]}<{primitive_type}>"
            elif len(shape) == 2:
                # matNxM is Matrix of N columns and M rows
                wgsl_type = f"mat{shape[1]}x{shape[0]}<{primitive_type}>"
            else:
                raise TypeError("Unsupported type {dtype}")
            # Check alignment (https://www.w3.org/TR/WGSL/#alignment-and-size)
            if wgsl_type == primitive_type:
                alignment = 4
            elif wgsl_type.startswith("vec"):
                c = int(wgsl_type.split("<")[0][-1])
                alignment = 8 if c < 3 else 16
            elif wgsl_type.startswith("mat"):
                c = int(wgsl_type.split("<")[0][-1])
                alignment = 8 if c < 3 else 16
            else:
                raise TypeError(f"Unsupported wgsl type: {wgsl_type}")
            if offset % alignment != 0:
                raise TypeError(
                    f"Struct alignment error: {name}.{fieldname} alignment must be {alignment}"
                )

            code += f"\n            {fieldname}: {wgsl_type};"

        code += f"""
        }};

        [[group({bindgroup}), binding({index})]]
        var {name}: {structname};
        """
        self._uniform_codes[name] = code



from typing import Tuple

from ..linalg import Vector3, Matrix4, Quaternion, Spherical


# todo: maybe make an OrbitOrthoControls for ortho cameras, instead of this zoom param?


def get_screen_vectors_in_world_cords(
    center_world: Vector3, canvas_size: Tuple[float, float], camera: "Camera"
) -> Tuple[Vector3, Vector3]:
    """Given a reference center location (in 3D world coordinates)
    Get the vectors corresponding to the x and y direction in screen coordinates.
    These vectors are scaled so that they can simply be multiplied with the
    delta x and delta y.
    """
    center = center_world.clone().project(camera)
    pos1 = Vector3(100, 0, center.z).unproject(camera)
    pos2 = Vector3(0, 100, center.z).unproject(camera)
    pos1.multiply_scalar(0.02 / canvas_size[0])
    pos2.multiply_scalar(0.02 / canvas_size[1])
    return pos1, pos2  # now they're vecs, really


class OrbitControls:
    def __init__(
        self,
        eye: Vector3 = None,
        target: Vector3 = None,
        up: Vector3 = None,
        zoom_changes_distance=True,
        min_zoom: float = 0.0001,
    ) -> None:
        self.rotation = Quaternion()
        if eye is None:
            eye = Vector3(50.0, 50.0, 50.0)
        if target is None:
            target = Vector3()
        if up is None:
            up = Vector3(0.0, 1.0, 0.0)
        self.zoom_changes_distance = bool(zoom_changes_distance)
        self.zoom_value = 1
        self.min_zoom = min_zoom

        # State info used during a pan or rotate operation
        self._pan_info = None
        self._rotate_info = None

        # Temp objects (to avoid garbage collection)
        self._m = Matrix4()
        self._v = Vector3()
        self._origin = Vector3()
        self._orbit_up = Vector3(0, 1, 0)
        self._s = Spherical()

        # Initialize orientation
        self.look_at(eye, target, up)
        self._initial_distance = self.distance

    def look_at(self, eye: Vector3, target: Vector3, up: Vector3) -> "OrbitControls":
        self.distance = eye.distance_to(target)
        self.target = target
        self.up = up
        self.rotation.set_from_rotation_matrix(self._m.look_at(eye, target, up))
        self._up_quat = Quaternion().set_from_unit_vectors(self.up, self._orbit_up)
        self._up_quat_inv = self._up_quat.clone().inverse()
        return self

    def pan(self, vec3: Vector3) -> "OrbitControls":
        """Pan in 3D world coordinates."""
        self.target.add(vec3)
        return self

    def pan_start(
        self,
        pos: Tuple[float, float],
        canvas_size: Tuple[float, float],
        camera: "Camera",
    ) -> "OrbitControls":
        """Start a panning operation based (2D) screen coordinates."""
        vecx, vecy = get_screen_vectors_in_world_cords(self.target, canvas_size, camera)
        self._pan_info = {"last": pos, "vecx": vecx, "vecy": vecy}
        return self

    def pan_stop(self) -> "OrbitControls":
        self._pan_info = None
        return self

    def pan_move(self, pos: Tuple[float, float]) -> "OrbitControls":
        """Pan the center of rotation, based on a (2D) screen location. Call pan_start first."""
        if self._pan_info is None:
            return
        delta = tuple((pos[i] - self._pan_info["last"][i]) for i in range(2))
        self.pan(
            self._pan_info["vecx"]
            .clone()
            .multiply_scalar(-delta[0])
            .add_scaled_vector(self._pan_info["vecy"], +delta[1])
        )
        self._pan_info["last"] = pos
        return self

    def rotate(self, theta: float, phi: float) -> "OrbitControls":
        """Rotate using angles (in radians). theta and phi are also known
        as azimuth and elevation.
        """
        # offset
        self._v.set(0, 0, self.distance).apply_quaternion(self.rotation)
        # to neutral up
        self._v.apply_quaternion(self._up_quat)
        # to spherical
        self._s.set_from_vector3(self._v)
        # apply delta
        self._s.theta -= theta
        self._s.phi -= phi
        # clip
        self._s.make_safe()
        # back to cartesian
        self._v.set_from_spherical(self._s)
        # back to camera up
        self._v.apply_quaternion(self._up_quat_inv)
        # compute new rotation
        self.rotation.set_from_rotation_matrix(
            self._m.look_at(self._v, self._origin, self.up)
        )
        return self

    def rotate_start(
        self,
        pos: Tuple[float, float],
        canvas_size: Tuple[float, float],
        camera: "Camera",
    ) -> "OrbitControls":
        """Start a rotation operation based (2D) screen coordinates."""
        self._rotate_info = {"last": pos}
        return self

    def rotate_stop(self) -> "OrbitControls":
        self._rotate_info = None
        return self

    def rotate_move(
        self, pos: Tuple[float, float], speed: float = 0.0175
    ) -> "OrbitControls":
        """Rotate, based on a (2D) screen location. Call rotate_start first.
        The speed is 1 degree per pixel by default.
        """
        if self._rotate_info is None:
            return
        delta = tuple((pos[i] - self._rotate_info["last"][i]) * speed for i in range(2))
        self.rotate(*delta)
        self._rotate_info["last"] = pos
        return self

    def zoom(self, multiplier: float) -> "OrbitControls":
        self.zoom_value = max(self.min_zoom, float(multiplier) * self.zoom_value)
        if self.zoom_changes_distance:
            self.distance = self._initial_distance / self.zoom_value
        return self

    def get_view(self) -> Tuple[Vector3, Vector3, float]:
        self._v.set(0, 0, self.distance).apply_quaternion(self.rotation).add(
            self.target
        )
        zoom = 1 if self.zoom_changes_distance else self.zoom_value
        return self.rotation, self._v, zoom

    def update_camera(self, camera: "Camera") -> "OrbitControls":
        rot, pos, zoom = self.get_view()
        camera.rotation.copy(rot)
        camera.position.copy(pos)
        camera.zoom = zoom
        return self
